# Copyright (c) 2015 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import eventlet
from oslo_log import log as logging

from congress.datalog import arithmetic_solvers
from congress.dse import d6cage
from congress import harness
from congress.policy_engines import vm_placement
from congress.tests import base
from congress.tests import helper

LOG = logging.getLogger(__name__)

NREC_THEORY = 'non-recursive theory'


class TestEngine(base.TestCase):

    def test_parse(self):
        engine = vm_placement.ComputePlacementEngine()
        engine.debug_mode()
        f = engine.parse1('nova:q(1)')
        self.assertTrue(f.table.table, 'nova:q')
        self.assertIsNone(f.table.service)

        f = engine.parse1('p(x) :- q(x)')
        self.assertEqual(f.head.table.table, 'p')
        self.assertEqual(f.body[0].table.table, 'q')

    def test_select(self):
        engine = vm_placement.ComputePlacementEngine()
        engine.debug_mode()
        engine.insert('p(x) :- q(x)')
        engine.insert('q(1)')
        ans = engine.select('p(x)')
        self.assertTrue(helper.datalog_equal(ans, 'p(1)'))

    def test_theory_in_head(self):
        engine = vm_placement.ComputePlacementEngine()
        engine.debug_mode()
        engine.policy.insert(engine.parse1('p(x) :- nova:q(x)'))
        engine.policy.insert(engine.parse1('nova:q(1)'))
        ans = engine.policy.select(engine.parse1('p(x)'))
        ans = " ".join(str(x) for x in ans)
        self.assertTrue(helper.datalog_equal(ans, 'p(1)'))


class TestSetPolicy(base.TestCase):
    """Tests for setting policy."""

    def setUp(self):
        # create DSE and add vm-placement engine and fake datasource
        super(TestSetPolicy, self).setUp()
        self.cage = d6cage.d6Cage()
        config = {'vmplace':
                  {'module': "congress/policy_engines/vm_placement.py"},
                  'fake':
                  {'poll_time': 0,
                   'module': "congress/tests/fake_datasource.py"}}

        harness.load_data_service("vmplace", config['vmplace'],
                                  self.cage, helper.root_path(), 1)
        harness.load_data_service("fake", config['fake'],
                                  self.cage, helper.root_path(), 2)

        self.vmplace = self.cage.service_object('vmplace')
        self.vmplace.debug_mode()
        self.fake = self.cage.service_object('fake')

    def test_set_policy_subscriptions(self):
        self.vmplace.set_policy('p(x) :- fake:q(x)')
        helper.retry_check_subscriptions(
            self.vmplace, [(self.fake.name, 'q')])
        helper.retry_check_subscribers(
            self.fake, [(self.vmplace.name, 'q')])

    def test_set_policy(self):
        LOG.info("set_policy")
        self.vmplace.set_policy('p(x) :- fake:q(x)')
        self.fake.state = {'q': set([tuple([1]), tuple([2])])}
        self.fake.poll()
        ans = ('p(1) p(2)')
        helper.retry_check_db_equal(self.vmplace, 'p(x)', ans)

    # TODO(thinrichs): add tests for data update
    #   Annoying since poll() saves self.state, invokes
    #   update_from_datasource (which updates self.state),
    #   computes deltas, and publishes.  No easy way to inject
    #   a new value for self.state and get it to send non-empty
    #   deltas over the message bus.  Probably want to extend
    #   fake_datasource to include a client (default to None), make
    #   update_from_datasource use that client to set self.state,
    #   and then mock out the client.

    # TODO(thinrichs): add tests for setting policy to something that
    #   requires tables to be unsubscribed from

    # TODO(thinrichs): test production_mode()


class TestLpLang(base.TestCase):
    """Test the DatalogLp language."""
    def test_variables(self):
        var1 = arithmetic_solvers.LpLang.makeVariable('alice', 1, 3.1)
        var2 = arithmetic_solvers.LpLang.makeVariable('alice', 1, 3.1)
        var3 = arithmetic_solvers.LpLang.makeVariable('alice', 1, 4.0)
        self.assertEqual(var1, var2)
        self.assertNotEqual(var1, var3)

    def test_or(self):
        var1 = arithmetic_solvers.LpLang.makeVariable('alice', 1)
        var2 = arithmetic_solvers.LpLang.makeVariable('bob', 1)
        var3 = arithmetic_solvers.LpLang.makeVariable('charlie', 1)
        p1 = arithmetic_solvers.LpLang.makeOr(var1, var2)
        p2 = arithmetic_solvers.LpLang.makeOr(var1, var2)
        p3 = arithmetic_solvers.LpLang.makeOr(var1, var3)
        self.assertEqual(p1, p2)
        self.assertNotEqual(p1, p3)
        p4 = arithmetic_solvers.LpLang.makeOr(var1)
        self.assertEqual(p4, var1)
        p5 = arithmetic_solvers.LpLang.makeOr(var2, var1)
        self.assertEqual(p1, p5)

    def test_and(self):
        var1 = arithmetic_solvers.LpLang.makeVariable('alice', 1)
        var2 = arithmetic_solvers.LpLang.makeVariable('bob', 1)
        var3 = arithmetic_solvers.LpLang.makeVariable('charlie', 1)
        p1 = arithmetic_solvers.LpLang.makeAnd(var1, var2)
        p2 = arithmetic_solvers.LpLang.makeAnd(var1, var2)
        p3 = arithmetic_solvers.LpLang.makeAnd(var1, var3)
        self.assertEqual(p1, p2)
        self.assertNotEqual(p1, p3)
        p4 = arithmetic_solvers.LpLang.makeAnd(var1)
        self.assertEqual(p4, var1)
        p5 = arithmetic_solvers.LpLang.makeAnd(var2, var1)
        self.assertEqual(p1, p5)

    def test_equal(self):
        var1 = arithmetic_solvers.LpLang.makeVariable('alice', 1)
        var2 = arithmetic_solvers.LpLang.makeVariable('bob', 1)
        var3 = arithmetic_solvers.LpLang.makeVariable('charlie', 1)
        p1 = arithmetic_solvers.LpLang.makeEqual(var1, var2)
        p2 = arithmetic_solvers.LpLang.makeEqual(var1, var2)
        p3 = arithmetic_solvers.LpLang.makeEqual(var1, var3)
        self.assertEqual(p1, p2)
        self.assertNotEqual(p1, p3)
        p4 = arithmetic_solvers.LpLang.makeEqual(var2, var1)
        self.assertEqual(p1, p4)

    def test_notequal(self):
        var1 = arithmetic_solvers.LpLang.makeVariable('alice', 1)
        var2 = arithmetic_solvers.LpLang.makeVariable('bob', 1)
        var3 = arithmetic_solvers.LpLang.makeVariable('charlie', 1)
        p1 = arithmetic_solvers.LpLang.makeNotEqual(var1, var2)
        p2 = arithmetic_solvers.LpLang.makeNotEqual(var1, var2)
        p3 = arithmetic_solvers.LpLang.makeNotEqual(var1, var3)
        self.assertEqual(p1, p2)
        self.assertNotEqual(p1, p3)
        p4 = arithmetic_solvers.LpLang.makeNotEqual(var2, var1)
        self.assertEqual(p1, p4)

    def test_arith(self):
        var1 = arithmetic_solvers.LpLang.makeVariable('alice', 1)
        var2 = arithmetic_solvers.LpLang.makeVariable('bob', 1)
        p1 = arithmetic_solvers.LpLang.makeArith('lt', var1, var2)
        p2 = arithmetic_solvers.LpLang.makeArith('lt', var1, var2)
        p3 = arithmetic_solvers.LpLang.makeArith('gt', var1, var2)
        self.assertEqual(p1, p2)
        self.assertNotEqual(p1, p3)

    def test_complex(self):
        var1 = arithmetic_solvers.LpLang.makeVariable('alice', 1)
        var2 = arithmetic_solvers.LpLang.makeVariable('bob', 1)
        arith1 = arithmetic_solvers.LpLang.makeArith('lt', var1, var2)
        arith2 = arithmetic_solvers.LpLang.makeArith('lt', var1, var2)
        arith3 = arithmetic_solvers.LpLang.makeArith('gt', var1, var2)
        p1 = arithmetic_solvers.LpLang.makeEqual(
            var1, arithmetic_solvers.LpLang.makeOr(arith1, arith2))
        p2 = arithmetic_solvers.LpLang.makeEqual(
            var1, arithmetic_solvers.LpLang.makeOr(arith2, arith1))
        p3 = arithmetic_solvers.LpLang.makeEqual(
            var1, arithmetic_solvers.LpLang.makeOr(arith1, arith3))

        # equality
        self.assertEqual(p1, p2)
        self.assertNotEqual(p1, p3)

        # sets
        s1 = set([arith1, p1])
        s2 = set([p1, arith1])
        s3 = set([arith1, arith2])
        self.assertEqual(s1, s2)
        self.assertNotEqual(s1, s3)


class TestDatalogToLp(base.TestCase):
    def check(self, code, data, query, ans, possibility=None):
        if possibility is None:
            possibility = []
        engine = vm_placement.ComputePlacementEngine(
            arithmetic_solvers.LpLang())
        engine.debug_mode()
        engine.insert(code)
        for d in data:
            engine.insert(d)
        query = engine.parse1(query)
        (rules, variables) = engine.datalog_to_lp(query, possibility)
        LOG.info("Checking equality")
        if not same_sets(rules, ans):
            LOG.info("-- actual --")
            for rule in rules:
                LOG.info("%s", rule)
            LOG.info("-- correct --")
            for rule in ans:
                LOG.info("%s", rule)
            self.fail("actual and correct mismatch")

    def test_basic(self):
        code = ('warning(id) :- '
                ' nova:host(id, zone, memory_capacity), '
                ' legacy:special_zone(zone), '
                ' ceilometer:mem_consumption(id, avg), '
                ' mul(0.75, memory_capacity, three_quarters_mem),'
                ' lt(avg, three_quarters_mem)')
        data = ('nova:host(123, "dmz", 10) ',
                'legacy:special_zone("dmz") ',
                'ceilometer:mem_consumption(123, 15)')  # ignored
        query = 'warning(x)'
        ans = arithmetic_solvers.LpLang.makeExpr(['eq',
                                                  ['var', 'warning', 123],
                                                  ['lt',
                                                   ['var', 'hMemUse', 123],
                                                   7.5]])
        self.check(code, data, query, [ans])

    def test_multiple_rows(self):
        code = ('warning(id) :- '
                ' nova:host(id, zone, memory_capacity), '
                ' legacy:special_zone(zone), '
                ' ceilometer:mem_consumption(id, avg), '
                ' mul(0.75, memory_capacity, three_quarters_mem),'
                ' lt(avg, three_quarters_mem)')
        data = ('nova:host(123, "dmz", 10) ',
                'nova:host(456, "dmz", 20) ',
                'legacy:special_zone("dmz") ',
                'ceilometer:mem_consumption(123, 15)')   # ignored
        query = 'warning(x)'
        ans1 = arithmetic_solvers.LpLang.makeExpr(['eq',
                                                   ['var', 'warning', 123],
                                                   ['lt',
                                                    ['var', 'hMemUse', 123],
                                                    7.5]])
        ans2 = arithmetic_solvers.LpLang.makeExpr(['eq',
                                                   ['var', 'warning', 456],
                                                   ['lt',
                                                    ['var', 'hMemUse', 456],
                                                    15.0]])
        self.check(code, data, query, [ans1, ans2])

    # def test_disjunction(self):
    #     code = ('warning(id) :- '
    #             ' nova:host(id, zone, memory_capacity), '
    #             ' legacy:special_zone(zone), '
    #             ' ceilometer:mem_consumption(id, avg), '
    #             ' mul(0.75, memory_capacity, three_quarters_mem),'
    #             ' lt(avg, three_quarters_mem)')
    #     data = ('nova:host(123, "dmz", 10) ',
    #             'nova:host(456, "dmz", 20) ',
    #             'nova:host(456, "dmz", 30) ',  # doesn't really make sense
    #             'legacy:special_zone("dmz") ',
    #             'ceilometer:mem_consumption(123, 15)')   # ignored
    #     query = 'warning(x)'
    #     ans1 = LpLang.makeExpr(['eq',
    #                             ['var', 'warning', 123],
    #                             ['lt', ['var', 'hMemUse', 123], 7.5]])
    #     ans2 = LpLang.makeExpr(['eq',
    #                             ['var', 'warning', 456],
    #                             ['or',
    #                              ['lt', ['var', 'hMemUse', 123], 7.5],
    #                              ['lt', ['var', 'hMemUse', 456], 15.0]]])
    #     self.check(code, data, query, [ans1, ans2])

    def test_multiple_constraints(self):
        code = ('warning(id) :- '
                ' nova:host(id, zone, memory_capacity), '
                ' legacy:special_zone(zone), '
                ' ceilometer:mem_consumption(id, avg), '
                ' mul(0.75, memory_capacity, three_quarters_mem),'
                ' lt(avg, three_quarters_mem),'
                ' lt(avg, 100)')
        data = ('nova:host(123, "dmz", 10) ',
                'legacy:special_zone("dmz") ',
                'ceilometer:mem_consumption(123, 15)')   # ignored
        query = 'warning(x)'
        ans1 = arithmetic_solvers.LpLang.makeExpr(
            ['eq',
             ['var', 'warning', 123],
             ['and', ['lt', ['var', 'hMemUse', 123], 7.5],
                     ['lt', ['var', 'hMemUse', 123], 100]]])
        self.check(code, data, query, [ans1])


class TestDatalogPolicyToLp(base.TestCase):
    def check(self, actual, correct):
        extra = diff(actual, correct)
        missing = diff(correct, actual)
        if len(extra) or len(missing):
            LOG.info("-- missing --")
            for rule in missing:
                LOG.info("%s", rule)
            LOG.info("-- extra --")
            for rule in extra:
                LOG.info("%s", rule)
            self.fail("actual and correct mismatch")

    def test_domain_axioms(self):
        engine = vm_placement.ComputePlacementEngine()
        engine.debug_mode()
        engine.insert('nova:host(123, 1, 10)')
        engine.insert('nova:host(456, 1, 10)')
        engine.insert('nova:server(789, "alice", 123)')
        engine.insert('nova:server(101, "bob", 123)')
        engine.insert('ceilometer:mem_consumption(789, 10)')
        engine.insert('ceilometer:mem_consumption(101, 20)')
        self.assertEqual(set(engine.get_hosts()), set([123, 456]))
        self.assertEqual(set(engine.get_guests()), set([789, 101]))
        ans1 = arithmetic_solvers.LpLang.makeExpr(
            ['eq',
             ['var', 'hMemUse', 456],
             ['plus',
              ['times', ['var', 'assign', 101, 456], 20],
              ['times', ['var', 'assign', 789, 456], 10]]])
        ans2 = arithmetic_solvers.LpLang.makeExpr(
            ['eq',
             ['var', 'hMemUse', 123],
             ['plus',
              ['times', ['var', 'assign', 101, 123], 20],
              ['times', ['var', 'assign', 789, 123], 10]]])
        ans3 = arithmetic_solvers.LpLang.makeExpr(
            ['eq',
             1,
             ['plus',
              ['var', 'assign', 101, 123],
              ['var', 'assign', 101, 456]]])
        ans4 = arithmetic_solvers.LpLang.makeExpr(
            ['eq',
             1,
             ['plus',
              ['var', 'assign', 789, 123],
              ['var', 'assign', 789, 456]]])
        self.check(engine.domain_axioms(), [ans1, ans2, ans3, ans4])

    def test_policy_to_lp(self):
        engine = vm_placement.ComputePlacementEngine()
        engine.debug_mode()
        engine.insert('nova:host(123, 1, 10)')
        engine.insert('nova:host(456, 1, 10)')
        engine.insert('nova:server(789, "alice", 123)')
        engine.insert('nova:server(101, "bob", 123)')
        code = ('warning(id) :- '
                ' nova:host(id, zone, memory_capacity), '
                ' legacy:special_zone(zone), '
                ' ceilometer:mem_consumption(id, avg), '
                ' mul(0.75, memory_capacity, three_quarters_mem),'
                ' lt(avg, three_quarters_mem)')
        engine.insert(code)
        engine.insert('legacy:special_zone(1)')
        engine.insert('ceilometer:mem_consumption(123, 15)')
        engine.insert('ceilometer:mem_consumption(456, 20)')
        engine.insert('ceilometer:mem_consumption(789, 5)')
        engine.insert('ceilometer:mem_consumption(101, 10)')
        opt, constraints = engine.policy_to_lp()

        optans = arithmetic_solvers.LpLang.makeExpr(['or',
                                                     ['var', 'warning', 456],
                                                     ['var', 'warning', 123]])
        ans1 = arithmetic_solvers.LpLang.makeExpr(['eq',
                                                   ['var', 'warning', 123],
                                                   ['lt',
                                                    ['var', 'hMemUse', 123],
                                                    7.5]])
        ans2 = arithmetic_solvers.LpLang.makeExpr(['eq',
                                                   ['var', 'warning', 456],
                                                   ['lt',
                                                    ['var', 'hMemUse', 456],
                                                    7.5]])
        ans3 = arithmetic_solvers.LpLang.makeExpr(
            ['eq',
             ['var', 'hMemUse', 456],
             ['plus',
              ['times', ['var', 'assign', 101, 456], 10],
              ['times', ['var', 'assign', 789, 456], 5]]])
        ans4 = arithmetic_solvers.LpLang.makeExpr(
            ['eq',
             ['var', 'hMemUse', 123],
             ['plus',
              ['times', ['var', 'assign', 101, 123], 10],
              ['times', ['var', 'assign', 789, 123], 5]]])
        ans5 = arithmetic_solvers.LpLang.makeExpr(
            ['eq',
             1,
             ['plus',
              ['var', 'assign', 101, 123],
              ['var', 'assign', 101, 456]]])
        ans6 = arithmetic_solvers.LpLang.makeExpr(
            ['eq',
             1,
             ['plus',
              ['var', 'assign', 789, 123],
              ['var', 'assign', 789, 456]]])

        self.check([opt], [optans])
        self.check(constraints, [ans1, ans2, ans3, ans4, ans5, ans6])


class TestPureLp(base.TestCase):
    """Test conversion of Datalog LP to pure LP."""
    def check(self, expr, bounds, correct):
        lang = arithmetic_solvers.LpLang()
        actual = lang.pure_lp(lang.makeExpr(expr), bounds)
        correct = [lang.makeExpr(x) for x in correct]
        if not same_sets(actual, correct):
            LOG.info("-- actual --")
            for rule in actual:
                LOG.info("%s", rule)
            LOG.info("-- correct --")
            for rule in correct:
                LOG.info("%s", rule)
            self.fail("actual and correct mismatch")

    def test_real_1(self):
        """Test a real use case."""
        exp = ['eq', ['var', 'warning', 123],
                     ['lteq', ['var', 'hMemUse', 123], 7.5]]
        bounds = {('VAR', 'hMemUse', 123): 100}
        y = ['var', 'warning', 123]
        x = ['minus',
             ['minus', ['var', 'hMemUse', 123], 7.5],
             arithmetic_solvers.LpLang.MIN_THRESHOLD]
        upperx = 101
        c1 = ['lteq', ['times', -1, x], ['times', y, upperx]]
        c2 = ['lt', x, ['times', ['minus', 1, y], upperx]]
        self.check(exp, bounds, [c1, c2])

    def test_real_2(self):
        e = ['eq',
             ['var', 'hMemUse', 456],
             ['plus',
              ['times', ['var', 'assign', 101, 456], ['var', 'gMemUse', 101]],
              ['times',
               ['var', 'assign', 789, 456],
               ['var', 'gMemUse', 789]]]]
        ans1 = ['eq',
                ['var', 'internal', 0],
                ['times',
                 ['var', 'assign', 101, 456],
                 ['var', 'gMemUse', 101]]]
        ans2 = ['eq',
                ['var', 'internal', 1],
                ['times',
                 ['var', 'assign', 789, 456],
                 ['var', 'gMemUse', 789]]]
        ans3 = ['eq',
                ['var', 'hMemUse', 456],
                ['plus', ['var', 'internal', 0], ['var', 'internal', 1]]]
        self.check(e, {}, [ans1, ans2, ans3])


class TestFlatten(base.TestCase):
    """Test reformulation of embedded operators into flattened formulas."""
    def check(self, input_expression, correct, correct_support):
        lang = arithmetic_solvers.LpLang()
        input_expr = lang.makeExpr(input_expression)
        LOG.info("input_expression: %s", input_expr)
        actual, actual_support = lang.flatten(input_expr, indicator=False)
        correct = lang.makeExpr(correct)
        correct_support = [lang.makeExpr(x) for x in correct_support]
        if (actual != correct or not same_sets(
                actual_support, correct_support)):
            LOG.info("-- actual: %s", actual)
            LOG.info("-- actual support --")
            for rule in actual_support:
                LOG.info("%s", rule)
            LOG.info("-- correct: %s", correct)
            LOG.info("-- correct support --")
            for rule in correct_support:
                LOG.info("%s", rule)
            self.fail("actual and correct mismatch")

    def test_flat(self):
        self.check(['or', 1, 2], ['or', 1, 2], [])
        self.check(['and', 1, 2], ['and', 1, 2], [])

    def test_nested(self):
        self.check(['or', 1, ['or', 2, 3]],  # orig
                   ['or', 1, ['var', 'internal', 0]],  # flat
                   [['eq', ['var', 'internal', 0],  # support
                           ['or', 2, 3]]])
        self.check(['or', 1, ['and', 2, 3]],  # orig
                   ['or', 1, ['var', 'internal', 0]],  # flat
                   [['eq', ['var', 'internal', 0],  # support
                           ['and', 2, 3]]])
        self.check(['eq',      # orig
                    ['var', 1, 2],
                    ['or',
                     ['and', 3, 4],
                     ['and', 5, 6]]],
                   ['eq',      # flat
                    ['var', 1, 2],
                    ['or', ['var', 'internal', 0], ['var', 'internal', 1]]],
                   [['eq', ['var', 'internal', 0], ['and', 3, 4]],  # support
                    ['eq', ['var', 'internal', 1], ['and', 5, 6]]])

    def test_real(self):
        self.check(['eq', ['var', 'warning', 123],
                          ['lt', ['var', 'hMemUse', 123], 7.5]],
                   ['eq', ['var', 'warning', 123],
                          ['lt', ['var', 'hMemUse', 123], 7.5]],
                   [])
        self.check(
            ['eq',  # orig
             ['var', 'hMemUse', 456],
             ['or',
              ['and', ['var', 'assign', 101, 456], ['var', 'gMemUse', 101]],
              ['and', ['var', 'assign', 789, 456], ['var', 'gMemUse', 789]]]],
            ['eq',  # flat
             ['var', 'hMemUse', 456],
             ['or', ['var', 'internal', 0], ['var', 'internal', 1]]],
            [['eq',  # support
              ['var', 'internal', 0],
              ['and', ['var', 'assign', 101, 456], ['var', 'gMemUse', 101]]],
             ['eq',
              ['var', 'internal', 1],
              ['and', ['var', 'assign', 789, 456], ['var', 'gMemUse', 789]]]])


class TestIndicatorElim(base.TestCase):
    """Test binary indicator variable elimination."""
    def check(self, input_expression, bounds, correct_expressions):
        lang = arithmetic_solvers.LpLang()
        input_expr = lang.makeExpr(input_expression)
        LOG.info("input_expression: %s", input_expr)
        actual = lang.indicator_to_pure_lp(input_expr, bounds)
        correct = [lang.makeExpr(x) for x in correct_expressions]
        extra = diff(actual, correct)
        missing = diff(correct, actual)
        if len(extra) or len(missing):
            LOG.info("-- missing --")
            for rule in missing:
                LOG.info("%s", rule)
            LOG.info("-- extra --")
            for rule in extra:
                LOG.info("%s", rule)
            self.fail("actual and correct mismatch")

    def test_simple(self):
        exp = ['eq', ['var', 'warning', 123],
                     ['lteq', ['var', 'hMemUse', 123], 7.5]]
        bounds = {('VAR', 'hMemUse', 123): 100}
        y = ['var', 'warning', 123]
        x = ['minus', ['minus', ['var', 'hMemUse', 123], 7.5],
             arithmetic_solvers.LpLang.MIN_THRESHOLD]
        upperx = 101
        c1 = ['lteq', ['times', -1, x], ['times', y, upperx]]
        c2 = ['lt', x, ['times', ['minus', 1, y], upperx]]
        self.check(exp, bounds, [c1, c2])


class TestToLtZero(base.TestCase):
    """Test conversion of inequality to form A < 0."""
    small = arithmetic_solvers.LpLang.MIN_THRESHOLD

    def check(self, expr, correct):
        lang = arithmetic_solvers.LpLang()
        actual = lang.arith_to_lt_zero(lang.makeExpr(expr))
        self.assertEqual(actual, lang.makeExpr(correct))

    def test_lt(self):
        expr = ['lt', 7, 8]
        self.check(expr, ['lt', ['minus', 7, 8], 0])

    def test_lteq(self):
        expr = ['lteq', 10, 11]
        self.check(expr, ['lt', ['minus', ['minus', 10, 11], self.small], 0])

    def test_gt(self):
        expr = ['gt', 12, 13]
        self.check(expr, ['lt', ['minus', 13, 12], 0])

    def test_gteq(self):
        expr = ['gteq', 14, 15]
        self.check(expr, ['lt', ['minus', ['minus', 15, 14], self.small], 0])


class TestUpperBound(base.TestCase):
    """Test upper bound computation."""
    def check(self, expr, bounds, correct):
        lang = arithmetic_solvers.LpLang()
        actual = lang.upper_bound(lang.makeExpr(expr), bounds)
        self.assertEqual(actual, correct)

    def test_times(self):
        exp = ['times', ['VAR', 'a', 1], ['VAR', 'a', 2]]
        bounds = {('VAR', 'a', 1): 2, ('VAR', 'a', 2): 3}
        self.check(exp, bounds, 6)

    def test_plus(self):
        exp = ['plus', ['VAR', 'a', 1], ['VAR', 'a', 2]]
        bounds = {('VAR', 'a', 1): 2, ('VAR', 'a', 2): 3}
        self.check(exp, bounds, 5)

    def test_minus(self):
        exp = ['minus', ['VAR', 'a', 1], ['VAR', 'a', 2]]
        bounds = {('VAR', 'a', 1): 2, ('VAR', 'a', 2): 3}
        self.check(exp, bounds, 2)

    def test_nested(self):
        exp = ['plus',
               ['times', ['VAR', 'a', 1], ['VAR', 'a', 2]],
               ['minus', ['VAR', 'a', 3], ['VAR', 'a', 4]]]
        bounds = {('VAR', 'a', 1): 5,
                  ('VAR', 'a', 2): 7,
                  ('VAR', 'a', 3): 2,
                  ('VAR', 'a', 4): 1}
        self.check(exp, bounds, 37)

    def test_nested2(self):
        exp = ['times',
               ['times', ['VAR', 'a', 1], ['VAR', 'a', 2]],
               ['minus', ['VAR', 'a', 3], ['VAR', 'a', 4]]]
        bounds = {('VAR', 'a', 1): 5,
                  ('VAR', 'a', 2): 7,
                  ('VAR', 'a', 3): 2,
                  ('VAR', 'a', 4): 1}
        self.check(exp, bounds, 70)


class TestComputeVmAssignment(base.TestCase):
    """Test full computation of VM assignment."""
    def test_two_servers(self):
        engine = vm_placement.ComputePlacementEngine()
        engine.debug_mode()
        engine.insert('nova:host(123, 1, 10)')
        engine.insert('nova:host(456, 1, 5)')
        engine.insert('nova:server(789, "alice", 123)')
        engine.insert('nova:server(101, "bob", 123)')
        code = ('warning(id) :- '
                ' nova:host(id, zone, memory_capacity), '
                ' legacy:special_zone(zone), '
                ' ceilometer:mem_consumption(id, avg), '
                ' mul(0.75, memory_capacity, three_quarters_mem),'
                ' lteq(avg, three_quarters_mem)')
        engine.insert(code)
        engine.insert('legacy:special_zone(1)')
        engine.insert('ceilometer:mem_consumption(789, 2)')
        engine.insert('ceilometer:mem_consumption(101, 2)')
        ans = engine.calculate_vm_assignment()
        LOG.info("assignment: %s", ans)
        self.assertEqual(ans, {101: 456, 789: 456})

    def test_three_servers(self):
        engine = vm_placement.ComputePlacementEngine()
        engine.debug_mode()
        engine.insert('nova:host(100, 1, 6)')
        engine.insert('nova:host(101, 1, 10)')
        engine.insert('nova:server(200, "alice", 100)')
        engine.insert('nova:server(201, "bob", 100)')
        engine.insert('nova:server(202, "bob", 101)')
        code = ('warning(id) :- '
                ' nova:host(id, zone, memory_capacity), '
                ' legacy:special_zone(zone), '
                ' ceilometer:mem_consumption(id, avg), '
                ' mul(0.75, memory_capacity, three_quarters_mem),'
                ' lteq(avg, three_quarters_mem)')
        engine.insert(code)
        engine.insert('legacy:special_zone(1)')
        engine.insert('ceilometer:mem_consumption(200, 2)')
        engine.insert('ceilometer:mem_consumption(201, 2)')
        engine.insert('ceilometer:mem_consumption(202, 2)')
        ans = engine.calculate_vm_assignment()
        LOG.info("assignment: %s", ans)
        self.assertEqual({200: 100, 201: 100, 202: 100}, ans)

    def test_set_policy(self):
        engine = vm_placement.ComputePlacementEngine(inbox=eventlet.Queue(),
                                                     datapath=eventlet.Queue())
        engine.debug_mode()
        p = (
            'nova:host(100, 1, 6)'
            'nova:host(101, 1, 10)'
            'nova:server(200, "alice", 100)'
            'nova:server(201, "bob", 100)'
            'nova:server(202, "bob", 101)'
            'warning(id) :- '
            ' nova:host(id, zone, memory_capacity), '
            ' legacy:special_zone(zone), '
            ' ceilometer:mem_consumption(id, avg), '
            ' mul(0.75, memory_capacity, three_quarters_mem),'
            ' lteq(avg, three_quarters_mem)'
            'legacy:special_zone(1)'
            'ceilometer:mem_consumption(100, 4)'
            'ceilometer:mem_consumption(101, 2)'
            'ceilometer:mem_consumption(200, 2)'
            'ceilometer:mem_consumption(201, 2)'
            'ceilometer:mem_consumption(202, 2)')
        engine.set_policy(p)
        LOG.info("assignment: %s", engine.guest_host_assignment)
        self.assertEqual({200: 100, 201: 100, 202: 100},
                         engine.guest_host_assignment)

    def test_set_policy_with_dashes(self):
        engine = vm_placement.ComputePlacementEngine(inbox=eventlet.Queue(),
                                                     datapath=eventlet.Queue())
        engine.debug_mode()
        p = (
            'nova:host("Server-100", 1, 6)'
            'nova:host("Server-101", 1, 10)'
            'nova:server(200, "alice", "Server-100")'
            'nova:server(201, "bob", "Server-100")'
            'nova:server(202, "bob", "Server-101")'
            'warning(id) :- '
            ' nova:host(id, zone, memory_capacity), '
            ' legacy:special_zone(zone), '
            ' ceilometer:mem_consumption(id, avg), '
            ' mul(0.75, memory_capacity, three_quarters_mem),'
            ' lteq(avg, three_quarters_mem)'
            'legacy:special_zone(1)'
            'ceilometer:mem_consumption("Server-100", 4)'
            'ceilometer:mem_consumption("Server-101", 2)'
            'ceilometer:mem_consumption(200, 2)'
            'ceilometer:mem_consumption(201, 2)'
            'ceilometer:mem_consumption(202, 2)')
        engine.set_policy(p)
        LOG.info("assignment: %s", engine.guest_host_assignment)
        self.assertEqual({200: 'Server-100', 201: 'Server-100',
                          202: 'Server-100'},
                         engine.guest_host_assignment)


def diff(iter1, iter2):
    iter1 = list(iter1)
    iter2 = list(iter2)
    ans = []
    for x in iter1:
        if x not in iter2:
            ans.append(x)
    return ans


def same_sets(iter1, iter2):
    """Can't use set(iter1) == set(iter2) b/c hash doesn't respect OR."""
    iter1 = list(iter1)  # make sure iter1 and iter2 are not Python sets
    iter2 = list(iter2)
    for x in iter1:
        if x not in iter2:
            return False
    for x in iter2:
        if x not in iter1:
            return False
    return True



# Copyright (c) 2015 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# This script starts a number of children processes defined in a config file.
# If restart_delay is not None, this will restart a crashed process after
# restart_delay seconds.  The config file content should look like this:
# {
#     "datasource_drivers":
#       [
#           { "cmd": "ls",
#             "args": ["-l"],
#             "restart_delay": 5,
#             "name": "foo"
#           },
#
#           { "cmd": "watch",
#             "args": ["ls", "-l", "/tmp"],
#             "name": "bar"
#           }
#       ],
#     "output_directory": "/tmp/runner"
# }
#
# Config file requirements: the "name" fields must be unique within a config
# file.
#
import json
import os
import signal
import subprocess
import sys
import time

from oslo_log import log as logging


LOG = logging.getLogger(__name__)

children = []


class ProcessEntry(object):
    def __init__(self, process, cmd, args, restart_delay, name):
        self.process = process
        self.cmd = cmd
        self.args = args
        self.restart_delay = restart_delay
        self.time_of_death = None
        self.name = name


def _pid_file_name(name):
    return name + '.pid'


def stop_process(name, output_dir):
    filename = _pid_file_name(name)
    if os.path.isfile(os.path.join(output_dir, filename)):
        try:
            f = open(os.path.join(output_dir, filename))
            pid = int(f.read().strip())
            LOG.debug("Killing process %s, pid %d", name, pid)
            os.kill(pid, signal.SIGKILL)
            f.close()
            os.unlink(filename)
        except ValueError:
            LOG.debug("Could not parse pid file: %s (process %s)", filename,
                      name)
        except OSError:
            LOG.debug("No such process %s, pid %d", pid)
        except IOError:
            LOG.debug("Failed to stop process %s, pid %d", name, pid)


def start_process(cmd, args, restart_delay, name, output_dir):
    out = open(os.path.join(output_dir, name + '.stdout'), 'w')
    err = open(os.path.join(output_dir, name + '.stderr'), 'w')

    LOG.debug("Starting process (" + name + "): " + cmd + ' ' + ' '.join(args))
    p = subprocess.Popen([cmd] + args, stdout=out, stderr=err)
    LOG.debug("Started as pid %d", p.pid)
    f = open(os.path.join(output_dir, _pid_file_name(name)), 'w')
    f.write('%d\n' % p.pid)
    f.close()
    children.append(ProcessEntry(p, cmd, args, restart_delay, name))


def wait_all(output_dir, poll_interval_ms):
    LOG.debug("Monitoring %d children", len(children))
    while(True):
        for c in children:
            c.process.poll()
            if c.process.returncode is not None:
                if c.time_of_death is None:
                    LOG.debug("pid %d ended at %s with return code %d, "
                              "process %s", c.process.pid, c.time_of_death,
                              c.process.returncode,
                              c.name)
                    c.time_of_death = time.time()
                if c.restart_delay is not None:
                    if c.time_of_death + c.restart_delay < time.time():
                        LOG.debug("Restarting " + c.cmd + ' ' +
                                  ' '.join(c.args))
                        children.remove(c)
                        start_process(c.cmd, c.args, c.restart_delay,
                                      c.name, output_dir)
                else:
                    children.remove(c)

        if not children:
            break
        time.sleep(poll_interval_ms/1000)


def main():
    if len(sys.argv) != 2:
        sys.stderr.write("usage: start_process.py config_file\n")
        sys.exit(1)

    with open(sys.argv[1]) as f:
        txt = f.read()
        f.close()
        config = json.loads(txt)

        if os.path.exists(config['output_directory']):
            if os.path.isfile(config['output_directory']):
                sys.stderr.write('output_directory %s already exists as '
                                 'a file\n', config['output_directory'])
                sys.exit(1)
        else:
            os.makedirs(config['output_directory'])

        names = set()
        for driver in config['datasource_drivers']:
            if driver['name'] in names:
                sys.stderr.write("Duplicate name '%s' in config file\n"
                                 % driver['name'])
                sys.exit(1)
            names.add(driver['name'])
            stop_process(driver['name'], config['output_directory'])

        for driver in config['datasource_drivers']:
            start_process(driver['cmd'], driver['args'],
                          driver.get('restart_delay'), driver['name'],
                          config['output_directory'])
        wait_all(config['output_directory'], config['poll_interval_ms'])


if __name__ == '__main__':
    main()



# Copyright 2014 VMware
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os

from oslo_config import cfg
from oslo_db import options as db_options
from oslo_log import log as logging
from oslo_middleware import cors
from oslo_policy import opts as policy_opts

from congress.managers import datasource as datasource_mgr
from congress import version

LOG = logging.getLogger(__name__)

core_opts = [
    cfg.StrOpt('bind_host', default='0.0.0.0',
               help="The host IP to bind to"),
    cfg.PortOpt('bind_port', default=1789,
                help="The port to bind to"),
    cfg.IntOpt('max_simultaneous_requests', default=1024,
               help="Thread pool size for eventlet."),
    cfg.BoolOpt('tcp_keepalive', default=False,
                help='Set this to true to enable TCP_KEEALIVE socket option '
                     'on connections received by the API server.'),
    cfg.IntOpt('tcp_keepidle',
               default=600,
               help='Sets the value of TCP_KEEPIDLE in seconds for each '
                    'server socket. Only applies if tcp_keepalive is '
                    'true. Not supported on OS X.'),
    cfg.StrOpt('policy_path',
               help="The path to the latest policy dump"),
    cfg.StrOpt('datasource_file',
               help="The file containing datasource configuration"),
    cfg.StrOpt('root_path',
               help="The absolute path to the congress repo"),
    cfg.IntOpt('api_workers', default=1,
               help='The number of worker processes to serve the congress '
                    'API application.'),
    cfg.StrOpt('api_paste_config', default="api-paste.ini",
               help=_("The API paste config file to use")),
    cfg.StrOpt('auth_strategy', default='keystone',
               help=_("The type of authentication to use")),
    cfg.ListOpt('drivers',
                default=[],
                help=_("List of driver class paths to import.")),
    cfg.IntOpt('datasource_sync_period', default=0,
               help='The number of seconds to wait between synchronizing '
                    'datasource config from the database '),
    cfg.BoolOpt('enable_execute_action', default=True,
                help="Sets the flag to False if you don't want the congress "
                "to execute actions."),
    cfg.BoolOpt('distributed_architecture', default=False,
                help="The flag to use congress new distributed architecture."
                "Don't set it to True in L release since the new architecture "
                "is under implementation."),
]

# Register the configuration options
cfg.CONF.register_opts(core_opts)

policy_opts.set_defaults(cfg.CONF, 'policy.json')
logging.register_options(cfg.CONF)

_SQL_CONNECTION_DEFAULT = 'sqlite://'
# Update the default QueuePool parameters. These can be tweaked by the
# configuration variables - max_pool_size, max_overflow and pool_timeout
db_options.set_defaults(cfg.CONF,
                        connection=_SQL_CONNECTION_DEFAULT,
                        sqlite_db='', max_pool_size=10,
                        max_overflow=20, pool_timeout=10)


def init(args, **kwargs):
    cfg.CONF(args=args, project='congress',
             version='%%(prog)s %s' % version.version_info.release_string(),
             **kwargs)
    datasource_mgr.DataSourceManager.validate_configured_drivers()


def setup_logging():
    """Sets up logging for the congress package."""
    logging.setup(cfg.CONF, 'congress')


def find_paste_config():
    config_path = cfg.CONF.find_file(cfg.CONF.api_paste_config)
    if not config_path:
        raise cfg.ConfigFilesNotFoundError(
            config_files=[cfg.CONF.api_paste_config])
    config_path = os.path.abspath(config_path)
    LOG.info(_("Config paste file: %s"), config_path)
    return config_path


def set_config_defaults():
    """This method updates all configuration default values."""
    # CORS Defaults
    # TODO(krotscheck): Update with https://review.openstack.org/#/c/285368/
    cfg.set_defaults(cors.CORS_OPTS,
                     allow_headers=['X-Auth-Token',
                                    'X-OpenStack-Request-ID',
                                    'X-Identity-Status',
                                    'X-Roles',
                                    'X-Service-Catalog',
                                    'X-User-Id',
                                    'X-Tenant-Id'],
                     expose_headers=['X-Auth-Token',
                                     'X-OpenStack-Request-ID',
                                     'X-Subject-Token',
                                     'X-Service-Token'],
                     allow_methods=['GET',
                                    'PUT',
                                    'POST',
                                    'DELETE',
                                    'PATCH']
                     )



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import mock

from congress.datasources import keystone_driver
from congress.tests import base
from congress.tests.datasources import util
from congress.tests import helper

ResponseObj = util.ResponseObj


class TestKeystoneDriver(base.TestCase):

    def setUp(self):
        super(TestKeystoneDriver, self).setUp()

        class FakeClient(object):
            def __init__(self):
                self.users = mock.MagicMock()
                self.roles = mock.MagicMock()
                self.tenants = mock.MagicMock()

        self.users_data = [
            ResponseObj({'username': 'alice',
                         'name': 'alice foo',
                         'enabled': True,
                         'tenantId': '019b18a15f2a44c1880d57704b2c4009',
                         'id': '00f2c34a156c40058004ee8eb3320e04',
                         'email': 'alice@foo.com'}),
            ResponseObj({'username': 'bob',
                         'name': 'bob bar',
                         'enabled': False,
                         'tenantId': 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',
                         'id': 'bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb',
                         'email': 'bob@bar.edu'})]

        self.roles_data = [
            ResponseObj({'id': 'cccccccccccccccccccccccccccccccc',
                         'name': 'admin'}),
            ResponseObj({'id': 'dddddddddddddddddddddddddddddddd',
                         'name': 'viewer'})]

        self.tenants_data = [
            ResponseObj({'enabled': True,
                         'description': 'accounting team',
                         'name': 'accounting',
                         'id': '00000000000000000000000000000001'}),
            ResponseObj({'enabled': False,
                         'description': 'eng team',
                         'name': 'eng',
                         'id': '00000000000000000000000000000002'})]

        self.keystone_client = mock.patch("keystoneclient.v2_0.client.Client",
                                          return_value=FakeClient())
        self.keystone_client.start()
        args = helper.datasource_openstack_args()
        args['poll_time'] = 0
        self.driver = keystone_driver.KeystoneDriver(args=args)
        self.driver.client.users.list.return_value = self.users_data
        self.driver.client.roles.list.return_value = self.roles_data
        self.driver.client.tenants.list.return_value = self.tenants_data

    def test_list_users(self):
        """Test conversion of complex user objects to tables."""
        self.driver.update_from_datasource()
        user_list = self.driver.state[keystone_driver.KeystoneDriver.USERS]
        self.assertIsNotNone(user_list)
        self.assertEqual(2, len(user_list))

        # Check an individual user entry
        self.assertTrue(('alice', 'alice foo', 'True',
                         '019b18a15f2a44c1880d57704b2c4009',
                         '00f2c34a156c40058004ee8eb3320e04',
                         'alice@foo.com') in user_list)
        self.assertTrue(('bob', 'bob bar', 'False',
                         'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',
                         'bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb',
                         'bob@bar.edu') in user_list)

    def test_list_roles(self):
        """Test conversion of complex role objects to tables."""
        self.driver.update_from_datasource()
        roles_list = self.driver.state[keystone_driver.KeystoneDriver.ROLES]
        self.assertIsNotNone(roles_list)
        self.assertEqual(2, len(roles_list))

        # Check an individual role entry
        self.assertTrue(('cccccccccccccccccccccccccccccccc', 'admin')
                        in roles_list)
        self.assertTrue(('dddddddddddddddddddddddddddddddd', 'viewer')
                        in roles_list)

    def test_list_tenants(self):
        """Test conversion of complex tenant objects to tables."""
        self.driver.update_from_datasource()
        tenants_key = keystone_driver.KeystoneDriver.TENANTS
        tenants_list = self.driver.state[tenants_key]
        self.assertIsNotNone(tenants_list)
        self.assertEqual(2, len(tenants_list))

        # Check an individual role entry
        self.assertTrue(('True', 'accounting team', 'accounting',
                         '00000000000000000000000000000001')
                        in tenants_list)
        self.assertTrue(('False', 'eng team', 'eng',
                         '00000000000000000000000000000002')
                        in tenants_list)

    def test_execute(self):
        class KeystoneClient(object):
            def __init__(self):
                self.testkey = None

            def enableProject(self, arg1):
                self.testkey = 'arg1=%s' % arg1

        keystone_client = KeystoneClient()
        self.driver.client = keystone_client
        api_args = {
            'positional': ['1']
        }
        expected_ans = 'arg1=1'

        self.driver.execute('enableProject', api_args)

        self.assertEqual(expected_ans, keystone_client.testkey)



# Copyright (c) 2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_config import cfg
cfg.CONF.distributed_architecture = True

from congress.api import policy_model
from congress.api import rule_model
from congress.api import table_model
from congress.api import webservice
from congress.tests import base
from congress.tests2.api import base as api_base


class TestTableModel(base.SqlTestCase):
    def setUp(self):
        super(TestTableModel, self).setUp()
        # Here we load the fake driver
        cfg.CONF.set_override(
            'drivers',
            ['congress.tests.fake_datasource.FakeDataSource'])

        self.table_model = table_model.TableModel('api-table',
                                                  policy_engine='engine')
        self.api_rule = rule_model.RuleModel('api-rule',
                                             policy_engine='engine')
        self.policy_model = policy_model.PolicyModel('api-policy',
                                                     policy_engine='engine')
        result = api_base.setup_config([self.table_model, self.api_rule,
                                        self.policy_model])
        self.node = result['node']
        self.engine = result['engine']
        self.data = result['data']
        # create test policy
        self._create_test_policy()

    def _create_test_policy(self):
        # create policy
        self.policy_model.add_item({"name": 'test-policy'}, {})

    def test_get_datasource_table_with_id(self):
        context = {'ds_id': self.data.service_id,
                   'table_id': 'fake_table'}
        expected_ret = {'id': 'fake_table'}
        ret = self.table_model.get_item('fake_table', {}, context)
        self.assertEqual(expected_ret, ret)

    def test_get_datasource_table_with_name(self):
        context = {'ds_id': self.data.service_id,
                   'table_id': 'fake_table'}
        expected_ret = {'id': 'fake_table'}
        ret = self.table_model.get_item('fake_table', {}, context)
        self.assertEqual(expected_ret, ret)

    def test_get_invalid_datasource(self):
        context = {'ds_id': 'invalid-id',
                   'table_id': 'fake_table'}
        self.assertRaises(webservice.DataModelException,
                          self.table_model.get_item, 'fake_table',
                          {}, context)

    def test_get_invalid_datasource_table(self):
        context = {'ds_id': self.data.service_id,
                   'table_id': 'invalid-table'}
        expected_ret = None
        ret = self.table_model.get_item('invalid-table', {}, context)
        self.assertEqual(expected_ret, ret)

    def test_get_policy_table(self):
        context = {'policy_id': 'test-policy',
                   'table_id': 'p'}
        expected_ret = {'id': 'p'}

        self.api_rule.add_item({'rule': 'p(x) :- q(x)'}, {}, context=context)
        self.api_rule.add_item({'rule': 'q(x) :- r(x)'}, {}, context=context)

        ret = self.table_model.get_item('p', {}, context)
        self.assertEqual(expected_ret, ret)

    def test_get_invalid_policy(self):
        context = {'policy_id': 'test-policy',
                   'table_id': 'fake-table'}
        invalid_context = {'policy_id': 'invalid-policy',
                           'table_id': 'fake-table'}
        expected_ret = None

        self.api_rule.add_item({'rule': 'p(x) :- q(x)'}, {}, context=context)
        self.api_rule.add_item({'rule': 'q(x) :- r(x)'}, {}, context=context)

        ret = self.table_model.get_item('test-policy',
                                        {}, invalid_context)
        self.assertEqual(expected_ret, ret)

    def test_get_invalid_policy_table(self):
        context = {'policy_id': 'test-policy',
                   'table_id': 'fake-table'}
        invalid_context = {'policy_id': 'test-policy',
                           'table_id': 'invalid-name'}
        expected_ret = None

        self.api_rule.add_item({'rule': 'p(x) :- q(x)'}, {}, context=context)
        self.api_rule.add_item({'rule': 'q(x) :- r(x)'}, {}, context=context)

        ret = self.table_model.get_item('test-policy', {},
                                        invalid_context)
        self.assertEqual(expected_ret, ret)

    def test_get_items_datasource_table(self):
        context = {'ds_id': self.data.service_id}
        expected_ret = {'results': [{'id': 'fake_table'}]}

        ret = self.table_model.get_items({}, context)
        self.assertEqual(expected_ret, ret)

    def test_get_items_invalid_datasource(self):
        context = {'ds_id': 'invalid-id',
                   'table_id': 'fake-table'}

        self.assertRaises(webservice.DataModelException,
                          self.table_model.get_items, {}, context)

    def _get_id_list_from_return(self, result):
        return [r['id'] for r in result['results']]

    def test_get_items_policy_table(self):
        context = {'policy_id': 'test-policy'}
        expected_ret = {'results': [{'id': x} for x in ['q', 'p', 'r']]}

        self.api_rule.add_item({'rule': 'p(x) :- q(x)'}, {}, context=context)
        self.api_rule.add_item({'rule': 'q(x) :- r(x)'}, {}, context=context)

        ret = self.table_model.get_items({}, context)
        self.assertEqual(set(self._get_id_list_from_return(expected_ret)),
                         set(self._get_id_list_from_return(ret)))

    def test_get_items_invalid_policy(self):
        context = {'policy_id': 'test-policy'}
        invalid_context = {'policy_id': 'invalid-policy'}
        expected_ret = None

        self.api_rule.add_item({'rule': 'p(x) :- q(x)'}, {}, context=context)
        self.api_rule.add_item({'rule': 'q(x) :- r(x)'}, {}, context=context)

        ret = self.table_model.get_items({}, invalid_context)
        self.assertEqual(expected_ret, ret)



# Copyright 2015 VMware.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from django.core.urlresolvers import reverse
from horizon import workflows
from openstack_dashboard.dashboards.admin.policies.rules import (
    workflows as rule_workflows)


LOG = logging.getLogger(__name__)


class CreateView(workflows.WorkflowView):
    workflow_class = rule_workflows.CreateRule
    ajax_template_name = 'admin/policies/rules/create.html'
    success_url = 'horizon:admin:policies:detail'

    def get_success_url(self):
        return reverse(self.success_url,
                       args=(self.kwargs['policy_name'],))

    def get_initial(self):
        return {'policy_name': self.kwargs['policy_name']}



# Copyright (c) 2015 Intel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

# set test to run as distributed arch
from oslo_config import cfg
cfg.CONF.distributed_architecture = True

from congress.api import action_model
from congress.api import webservice
from congress.tests import base
from congress.tests2.api import base as api_base


class TestActionModel(base.SqlTestCase):
    def setUp(self):
        super(TestActionModel, self).setUp()
        self.action_model = action_model.ActionsModel(
            'api-action', policy_engine='engine')
        services = api_base.setup_config([self.action_model])
        self.datasource = services['data']

    def test_get_datasource_actions(self):
        context = {'ds_id': self.datasource.service_id}
        actions = self.action_model.get_items({}, context=context)
        expected_ret = {'results': [{'name': 'fake_act',
                        'args': [{'name': 'server_id',
                                  'description': 'server to act'}],
                        'description': 'fake action'}]}
        self.assertEqual(expected_ret, actions)

    def test_get_invalid_datasource_action(self):
        context = {'ds_id': 'invalid_id'}
        self.assertRaises(webservice.DataModelException,
                          self.action_model.get_items, {}, context=context)



# Copyright (c) 2015 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging
import six
from six.moves import range

from congress.datalog import base
from congress.datalog.builtin import congressbuiltin
from congress.datalog import compile
from congress.datalog import unify
from congress.datalog import utility

LOG = logging.getLogger(__name__)


class TopDownTheory(base.Theory):
    """Class that holds the Top-Down evaluation routines.

    Classes will inherit from this class if they want to import and specialize
    those routines.
    """
    class TopDownContext(object):
        """Struct for storing the search state of top-down evaluation."""
        def __init__(self, literals, literal_index, binding, context, theory,
                     depth):
            self.literals = literals
            self.literal_index = literal_index
            self.binding = binding
            self.previous = context
            self.theory = theory   # a theory object, not just its name
            self.depth = depth

        def __str__(self):
            return (
                "TopDownContext<literals={}, literal_index={}, binding={}, "
                "previous={}, theory={}, depth={}>").format(
                    "[" + ",".join([str(x) for x in self.literals]) + "]",
                    str(self.literal_index), str(self.binding),
                    str(self.previous), self.theory.name, str(self.depth))

    class TopDownResult(object):
        """Stores a single result for top-down-evaluation."""
        def __init__(self, binding, support):
            self.binding = binding
            self.support = support   # for abduction

        def __str__(self):
            return "TopDownResult(binding={}, support={})".format(
                unify.binding_str(self.binding), utility.iterstr(self.support))

    class TopDownCaller(object):
        """Struct for info about the original caller of top-down evaluation.

        VARIABLES is the list of variables (from the initial query)
            that we want bindings for.
        BINDING is the initially empty BiUnifier.
        FIND_ALL controls whether just the first or all answers are found.
        ANSWERS is populated by top-down evaluation: it is the list of
               VARIABLES instances that the search process proved true.
        """

        def __init__(self, variables, binding, theory,
                     find_all=True, save=None):
            # an iterable of variable objects
            self.variables = variables
            # a bi-unifier
            self.binding = binding
            # the top-level theory (for included theories)
            self.theory = theory
            # a boolean
            self.find_all = find_all
            # The results of top-down-eval: a list of TopDownResults
            self.results = []
            # a Function that takes a compile.Literal and a unifier and
            #   returns T iff that literal under the unifier should be
            #   saved as part of an abductive explanation
            self.save = save
            # A variable used to store explanations as they are constructed
            self.support = []

        def __str__(self):
            return (
                "TopDownCaller<variables={}, binding={}, find_all={}, "
                "results={}, save={}, support={}>".format(
                    utility.iterstr(self.variables), str(self.binding),
                    str(self.find_all), utility.iterstr(self.results),
                    repr(self.save), utility.iterstr(self.support)))

    #########################################
    # External interface

    def __init__(self, name=None, abbr=None, theories=None, schema=None,
                 desc=None, owner=None):
        super(TopDownTheory, self).__init__(
            name=name, abbr=abbr, theories=theories, schema=schema,
            desc=desc, owner=owner)
        self.includes = []

    def select(self, query, find_all=True):
        """Return list of instances of QUERY that are true.

        If FIND_ALL is False, the return list has at most 1 element.
        """
        assert compile.is_datalog(query), "Query must be atom/rule"
        if compile.is_atom(query):
            literals = [query]
        else:
            literals = query.body
        # Because our output is instances of QUERY, need all the variables
        #   in QUERY.
        bindings = self.top_down_evaluation(query.variables(), literals,
                                            find_all=find_all)
        # LOG.debug("Top_down_evaluation returned: %s", bindings)
        if len(bindings) > 0:
            self.log(query.tablename(), "Found answer %s",
                     "[" + ",".join([str(query.plug(x))
                                    for x in bindings]) + "]")
        return [query.plug(x) for x in bindings]

    def explain(self, query, tablenames, find_all=True):
        """Return list of instances of QUERY that are true.

        Same as select except stores instances of TABLENAMES
        that participated in each proof. If QUERY is an atom,
        returns list of rules with QUERY in the head and
        the stored instances of TABLENAMES in the body; if QUERY is
        a rule, the rules returned have QUERY's head in the head
        and the stored instances of TABLENAMES in the body.
        """
        # This is different than abduction because instead of replacing
        #   a proof attempt with saving a literal, we want to save a literal
        #   after a successful proof attempt.
        assert False, "Not yet implemented"

    def abduce(self, query, tablenames, find_all=True):
        """Compute additional literals.

        Computes additional literals that if true would make
        (some instance of) QUERY true.  Returns a list of rules
        where the head represents an instance of the QUERY and
        the body is the collection of literals that must be true
        in order to make that instance true.  If QUERY is a rule,
        each result is an instance of the head of that rule, and
        the computed literals if true make the body of that rule
        (and hence the head) true.  If FIND_ALL is true, the
        return list has at most one element.
        Limitation: every negative literal relevant to a proof of
        QUERY is unconditionally true, i.e. no literals are saved
        when proving a negative literal is true.
        """
        assert compile.is_datalog(query), "abduce requires a formula"
        if compile.is_atom(query):
            literals = [query]
            output = query
        else:
            literals = query.body
            output = query.head
        # We need all the variables we will be using in the output, which
        #   here is just the head of QUERY (or QUERY itself if it is an atom)
        abductions = self.top_down_abduction(
            output.variables(), literals, find_all=find_all,
            save=lambda lit, binding: lit.tablename() in tablenames)
        results = [compile.Rule(output.plug(abd.binding), abd.support)
                   for abd in abductions]
        self.log(query.tablename(), "abduction result:")
        self.log(query.tablename(), "\n".join([str(x) for x in results]))
        return results

    def consequences(self, filter=None, table_theories=None):
        """Return all the true instances of any table in this theory."""
        # find all table, theory pairs defined in this theory
        if table_theories is None:
            table_theories = set()
            for key in self.rules.keys():
                table_theories |= set([(rule.head.table.table,
                                        rule.head.table.service)
                                       for rule in self.rules.get_rules(key)])
        results = set()
        # create queries: need table names and arities
        # TODO(thinrichs): arity computation will need to ignore
        #   modals once we start using insert[p(x)] instead of p+(x)
        for (table, theory) in table_theories:
            if filter is None or filter(table):
                tablename = compile.Tablename(table, theory)
                arity = self.arity(tablename)
                vs = []
                for i in range(0, arity):
                    vs.append("x" + str(i))
                vs = [compile.Variable(var) for var in vs]
                tablename = table
                if theory:
                    tablename = theory + ":" + tablename
                query = compile.Literal(tablename, vs)
                results |= set(self.select(query))
        return results

    def top_down_evaluation(self, variables, literals,
                            binding=None, find_all=True):
        """Compute bindings.

        Compute all bindings of VARIABLES that make LITERALS
        true according to the theory (after applying the unifier BINDING).
        If FIND_ALL is False, stops after finding one such binding.
        Returns a list of dictionary bindings.
        """
        # LOG.debug("CALL: top_down_evaluation(vars=%s, literals=%s, "
        #               "binding=%s)",
        #         ";".join(str(x) for x in variables),
        #         ";".join(str(x) for x in literals),
        #         str(binding))
        results = self.top_down_abduction(variables, literals,
                                          binding=binding, find_all=find_all,
                                          save=None)
        # LOG.debug("EXIT: top_down_evaluation(vars=%s, literals=%s, "
        #               "binding=%s) returned %s",
        #         iterstr(variables), iterstr(literals),
        #         str(binding), iterstr(results))
        return [x.binding for x in results]

    def top_down_abduction(self, variables, literals, binding=None,
                           find_all=True, save=None):
        """Compute bindings.

        Compute all bindings of VARIABLES that make LITERALS
        true according to the theory (after applying the
        unifier BINDING), if we add some number of additional
        literals.  Note: will not save any literals that are
        needed to prove a negated literal since the results
        would not make sense.  Returns a list of TopDownResults.
        """
        if binding is None:
            binding = self.new_bi_unifier()
        caller = self.TopDownCaller(variables, binding, self,
                                    find_all=find_all, save=save)
        if len(literals) == 0:
            self._top_down_finish(None, caller)
        else:
            # Note: must use same unifier in CALLER and CONTEXT
            context = self.TopDownContext(literals, 0, binding, None, self, 0)
            self._top_down_eval(context, caller)
        return list(set(caller.results))

    #########################################
    # Internal implementation

    def _top_down_eval(self, context, caller):
        """Compute instances.

        Compute all instances of LITERALS (from LITERAL_INDEX and above)
        that are true according to the theory (after applying the
        unifier BINDING to LITERALS).
        Returns True if done searching and False otherwise.
        """
        # no recursive rules, ever; this style of algorithm will not terminate
        lit = context.literals[context.literal_index]
        # LOG.debug("CALL: %s._top_down_eval(%s, %s)",
        #     self.name, context, caller)

        # abduction
        if caller.save is not None and caller.save(lit, context.binding):
            self._print_call(lit, context.binding, context.depth)
            # save lit and binding--binding may not be fully flushed out
            #   when we save (or ever for that matter)
            caller.support.append((lit, context.binding))
            self._print_save(lit, context.binding, context.depth)
            success = self._top_down_finish(context, caller)
            caller.support.pop()  # pop in either case
            if success:
                return True
            else:
                self._print_fail(lit, context.binding, context.depth)
                return False

        # regular processing
        if lit.is_negated():
            # LOG.debug("%s is negated", lit)
            # recurse on the negation of the literal
            plugged = lit.plug(context.binding)
            assert plugged.is_ground(), (
                "Negated literal not ground when evaluated: " +
                str(plugged))
            self._print_call(lit, context.binding, context.depth)
            new_context = self.TopDownContext(
                [lit.complement()], 0, context.binding, None,
                self, context.depth + 1)
            new_caller = self.TopDownCaller(caller.variables, caller.binding,
                                            caller.theory, find_all=False,
                                            save=None)
            # Make sure new_caller has find_all=False, so we stop as soon
            #    as we can.
            # Ensure save=None so that abduction does not save anything.
            #    Saving while performing NAF makes no sense.
            self._top_down_eval(new_context, new_caller)
            if len(new_caller.results) > 0:
                self._print_fail(lit, context.binding, context.depth)
                return False   # not done searching, b/c we failed
            else:
                # don't need bindings b/c LIT must be ground
                return self._top_down_finish(context, caller, redo=False)
        elif lit.tablename() == 'true':
            self._print_call(lit, context.binding, context.depth)
            return self._top_down_finish(context, caller, redo=False)
        elif lit.tablename() == 'false':
            self._print_fail(lit, context.binding, context.depth)
            return False
        elif lit.is_builtin():
            return self._top_down_builtin(context, caller)
        elif (self.theories is not None and
              lit.table.service is not None and
              lit.table.modal is None and  # not a modal
              lit.table.service != self.name and
              not lit.is_update()):  # not a pseudo-modal
            return self._top_down_module(context, caller)
        else:
            return self._top_down_truth(context, caller)

    def _top_down_builtin(self, context, caller):
        """Evaluate a table with a builtin semantics.

        Returns True if done searching and False otherwise.
        """
        lit = context.literals[context.literal_index]
        self._print_call(lit, context.binding, context.depth)
        builtin = congressbuiltin.builtin_registry.builtin(lit.table)
        # copy arguments into variables
        # PLUGGED is an instance of compile.Literal
        plugged = lit.plug(context.binding)
        # PLUGGED.arguments is a list of compile.Term
        # create args for function
        args = []
        for i in range(0, builtin.num_inputs):
            # save builtins with unbound vars during evaluation
            if not plugged.arguments[i].is_object() and caller.save:
                # save lit and binding--binding may not be fully flushed out
                #   when we save (or ever for that matter)
                caller.support.append((lit, context.binding))
                self._print_save(lit, context.binding, context.depth)
                success = self._top_down_finish(context, caller)
                caller.support.pop()  # pop in either case
                if success:
                    return True
                else:
                    self._print_fail(lit, context.binding, context.depth)
                    return False
            assert plugged.arguments[i].is_object(), (
                ("Builtins must be evaluated only after their "
                 "inputs are ground: {} with num-inputs {}".format(
                     str(plugged), builtin.num_inputs)))
            args.append(plugged.arguments[i].name)
        # evaluate builtin: must return number, string, or iterable
        #    of numbers/strings
        try:
            result = builtin.code(*args)
        except Exception as e:
            errmsg = "Error in builtin: " + str(e)
            self._print_note(lit, context.binding, context.depth, errmsg)
            self._print_fail(lit, context.binding, context.depth)
            return False

        # self._print_note(lit, context.binding, context.depth,
        #                 "Result: " + str(result))
        success = None
        undo = []
        if builtin.num_outputs > 0:
            # with return values, local success means we can bind
            #  the results to the return value arguments
            if (isinstance(result,
                           (six.integer_types, float, six.string_types))):
                result = [result]
            # Turn result into normal objects
            result = [compile.Term.create_from_python(x) for x in result]
            # adjust binding list
            unifier = self.new_bi_unifier()
            undo = unify.bi_unify_lists(result,
                                        unifier,
                                        lit.arguments[builtin.num_inputs:],
                                        context.binding)
            success = undo is not None
        else:
            # without return values, local success means
            #   result was True according to Python
            success = bool(result)

        if not success:
            self._print_fail(lit, context.binding, context.depth)
            unify.undo_all(undo)
            return False

        # otherwise, try to finish proof.  If success, return True
        if self._top_down_finish(context, caller, redo=False):
            unify.undo_all(undo)
            return True
        # if fail, return False.
        else:
            unify.undo_all(undo)
            self._print_fail(lit, context.binding, context.depth)
            return False

    def _top_down_module(self, context, caller):
        """Move to another theory and continue evaluation."""
        # LOG.debug("%s._top_down_module(%s)", self.name, context)
        lit = context.literals[context.literal_index]
        if lit.table.service not in self.theories:
            self._print_call(lit, context.binding, context.depth)
            errmsg = "No such policy: %s" % lit.table.service
            self._print_note(lit, context.binding, context.depth, errmsg)
            self._print_fail(lit, context.binding, context.depth)
            return False
        return self.theories[lit.table.service]._top_down_eval(context, caller)

    def _top_down_truth(self, context, caller):
        """Top down evaluation.

        Do top-down evaluation over the root theory at which
        the call was made and all the included theories.
        """
        # return self._top_down_th(context, caller)
        return self._top_down_includes(context, caller)

    def _top_down_includes(self, context, caller):
        """Top-down evaluation of all the theories included in this theory."""
        is_true = self._top_down_th(context, caller)
        if is_true and not caller.find_all:
            return True
        for th in self.includes:
            is_true = th._top_down_includes(context, caller)
            if is_true and not caller.find_all:
                return True
        return False

    def _top_down_th(self, context, caller):
        """Top-down evaluation for the rules in self."""
        # LOG.debug("%s._top_down_th(%s)", self.name, context)
        lit = context.literals[context.literal_index]
        self._print_call(lit, context.binding, context.depth)
        for rule in self.head_index(lit.table.table,
                                    lit.plug(context.binding)):
            unifier = self.new_bi_unifier()
            self._print_note(lit, context.binding, context.depth,
                             "Trying %s" % rule)
            # Prefer to bind vars in rule head
            undo = self.bi_unify(self.head(rule), unifier, lit,
                                 context.binding, self.name)
            if undo is None:  # no unifier
                continue
            if len(self.body(rule)) == 0:
                if self._top_down_finish(context, caller):
                    unify.undo_all(undo)
                    if not caller.find_all:
                        return True
                else:
                    unify.undo_all(undo)
            else:
                new_context = self.TopDownContext(
                    rule.body, 0, unifier, context, self, context.depth + 1)
                if self._top_down_eval(new_context, caller):
                    unify.undo_all(undo)
                    if not caller.find_all:
                        return True
                else:
                    unify.undo_all(undo)
        self._print_fail(lit, context.binding, context.depth)
        return False

    def _top_down_finish(self, context, caller, redo=True):
        """Helper function.

        This is called once top_down successfully completes
        a proof for a literal.  Handles (i) continuing search
        for those literals still requiring proofs within CONTEXT,
        (ii) adding solutions to CALLER once all needed proofs have
        been found, and (iii) printing out Redo/Exit during tracing.
        Returns True if the search is finished and False otherwise.
        Temporary, transparent modification of CONTEXT.
        """
        if context is None:
            # Found an answer; now store it
            if caller is not None:
                # flatten bindings and store before we undo
                # copy caller.support and store before we undo
                binding = {}
                for var in caller.variables:
                    binding[var] = caller.binding.apply(var)
                result = self.TopDownResult(
                    binding, [support[0].plug(support[1], caller=caller)
                              for support in caller.support])
                caller.results.append(result)
            return True
        else:
            self._print_exit(context.literals[context.literal_index],
                             context.binding, context.depth)
            # continue the search
            if context.literal_index < len(context.literals) - 1:
                context.literal_index += 1
                finished = context.theory._top_down_eval(context, caller)
                context.literal_index -= 1  # in case answer is False
            else:
                finished = self._top_down_finish(context.previous, caller)
            # return search result (after printing a Redo if failure)
            if redo and (not finished or caller.find_all):
                self._print_redo(context.literals[context.literal_index],
                                 context.binding, context.depth)
            return finished

    def _print_call(self, literal, binding, depth):
        msg = "{}Call: %s".format("| " * depth)
        self.log(literal.tablename(), msg, literal.plug(binding))

    def _print_exit(self, literal, binding, depth):
        msg = "{}Exit: %s".format("| " * depth)
        self.log(literal.tablename(), msg, literal.plug(binding))

    def _print_save(self, literal, binding, depth):
        msg = "{}Save: %s".format("| " * depth)
        self.log(literal.tablename(), msg, literal.plug(binding))

    def _print_fail(self, literal, binding, depth):
        msg = "{}Fail: %s".format("| " * depth)
        self.log(literal.tablename(), msg, literal.plug(binding))
        return False

    def _print_redo(self, literal, binding, depth):
        msg = "{}Redo: %s".format("| " * depth)
        self.log(literal.tablename(), msg, literal.plug(binding))
        return False

    def _print_note(self, literal, binding, depth, msg):
        self.log(literal.tablename(), "{}Note: {}".format("| " * depth,
                 msg))

    #########################################
    # Routines for specialization

    @classmethod
    def new_bi_unifier(cls, dictionary=None):
        """Return a unifier compatible with unify.bi_unify."""
        return unify.BiUnifier(dictionary=dictionary)
        # lambda (index):
        # compile.Variable("x" + str(index)), dictionary=dictionary)

    def defined_tablenames(self):
        """Returns list of table names defined in/written to this theory."""
        raise NotImplementedError

    def head_index(self, table, match_literal=None):
        """Return head index.

        This routine must return all the formulas pertinent for
        top-down evaluation when a literal with TABLE is at the top
        of the stack.
        """
        raise NotImplementedError

    def head(self, formula):
        """Given the output from head_index(), return the formula head.

        Given a FORMULA, return the thing to unify against.
        Usually, FORMULA is a compile.Rule, but it could be anything
        returned by HEAD_INDEX.
        """
        raise NotImplementedError

    def body(self, formula):
        """Return formula body.

        Given a FORMULA, return a list of things to push onto the
        top-down eval stack.
        """
        raise NotImplementedError

    def bi_unify(self, head, unifier1, body_element, unifier2, theoryname):
        """Unify atoms.

        Given something returned by self.head HEAD and an element in
        the return of self.body BODY_ELEMENT, modify UNIFIER1 and UNIFIER2
        so that HEAD.plug(UNIFIER1) == BODY_ELEMENT.plug(UNIFIER2).
        Returns changes that can be undone via unify.undo-all.
        THEORYNAME is the name of the theory for HEAD.
        """
        return unify.bi_unify_atoms(head, unifier1, body_element, unifier2,
                                    theoryname)

    #########################################
    # Routines for unknowns

    def instances(self, rule, possibilities=None):
        results = set([])
        possibilities = possibilities or []
        self._instances(rule, 0, self.new_bi_unifier(), results, possibilities)
        return results

    def _instances(self, rule, index, binding, results, possibilities):
        """Return all instances of the given RULE without evaluating builtins.

        Assumes self.head_index returns rules with empty bodies.
        """
        if index >= len(rule.body):
            results.add(rule.plug(binding))
            return
        lit = rule.body[index]
        self._print_call(lit, binding, 0)
        # if already ground or a builtin, go to the next literal
        if (lit.is_ground() or lit.is_builtin()):
            self._instances(rule, index + 1, binding, results, possibilities)
            return
        # Otherwise, find instances in this theory
        if lit.tablename() in possibilities:
            options = possibilities[lit.tablename()]
        else:
            options = self.head_index(lit.tablename(), lit.plug(binding))
        for data in options:
            self._print_note(lit, binding, 0, "Trying: %s" % repr(data))
            undo = unify.match_atoms(lit, binding, self.head(data))
            if undo is None:  # no unifier
                continue
            self._print_exit(lit, binding, 0)
            # recurse on the rest of the literals in the rule
            self._instances(rule, index + 1, binding, results, possibilities)
            if undo is not None:
                unify.undo_all(undo)
            self._print_redo(lit, binding, 0)
        self._print_fail(lit, binding, 0)



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging

from congress.api import api_utils
from congress.api import base
from congress.api import webservice
from congress import exception

LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    return TableModel(name, keys, inbox=inbox, dataPath=datapath, **args)


class TableModel(base.APIModel):
    """Model for handling API requests about Tables."""
    def __init__(self, name, keys='', inbox=None, dataPath=None,
                 policy_engine=None, datasource_mgr=None):
        super(TableModel, self).__init__(name, keys, inbox=inbox,
                                         dataPath=dataPath,
                                         policy_engine=policy_engine,
                                         datasource_mgr=datasource_mgr)

    def get_item(self, id_, params, context=None):
        """Retrieve item with id id_ from model.

        Args:
            id_: The ID of the item to retrieve
            params: A dict-like object containing parameters
                    from the request query string and body.
            context: Key-values providing frame of reference of request

        Returns:
             The matching item or None if item with id_ does not exist.
        """
        caller, source_id = api_utils.get_id_from_context(
            context,
            self.datasource_mgr,
            self.engine)

        args = {'source_id': source_id, 'table_id': id_}
        try:
            tablename = self.invoke_rpc(caller, 'get_tablename', args)
        except exception.CongressException as e:
            LOG.exception("Exception occurred while retrieving table %s"
                          "from datasource %s", id_, source_id)
            raise webservice.DataModelException.create(e)

        if tablename:
            return {'id': tablename}

        LOG.info('table id %s is not found in datasource %s', id_, source_id)

    def get_items(self, params, context=None):
        """Get items in model.

        Args:
            params: A dict-like object containing parameters
                    from the request query string and body.
            context: Key-values providing frame of reference of request

        Returns: A dict containing at least a 'results' key whose value is
                 a list of items in the model.  Additional keys set in the
                 dict will also be rendered for the user.
        """
        LOG.info('get_items has context %s', context)

        caller, source_id = api_utils.get_id_from_context(
            context,
            self.datasource_mgr,
            self.engine)

        try:
            tablenames = self.invoke_rpc(caller, 'get_tablenames',
                                         {'source_id': source_id})
        except exception.CongressException as e:
            LOG.exception("Exception occurred while retrieving tables"
                          "from datasource %s", source_id)
            raise webservice.DataModelException.create(e)
        # when the source_id doesn't have any table, 'tablenames' is set([])
        if isinstance(tablenames, set) or isinstance(tablenames, list):
            return {'results': [{'id': x} for x in tablenames]}

    # Tables can only be created/updated/deleted by writing policy
    #   or by adding new data sources.  Once we have internal data sources
    #   we need to implement all of these.

    # def add_item(self, item, id_=None, context=None):
    #     """Add item to model.

    #     Args:
    #         item: The item to add to the model
    #         id_: The ID of the item, or None if an ID should be generated
    #         context: Key-values providing frame of reference of request

    #     Returns:
    #          Tuple of (ID, newly_created_item)

    #     Raises:
    #         KeyError: ID already exists.
    #     """

    # def update_item(self, id_, item, context=None):
    #     """Update item with id_ with new data.

    #     Args:
    #         id_: The ID of the item to be updated
    #         item: The new item
    #         context: Key-values providing frame of reference of request

    #     Returns:
    #          The updated item.

    #     Raises:
    #         KeyError: Item with specified id_ not present.
    #     """
    #     # currently a noop since the owner_id cannot be changed
    #     if id_ not in self.items:
    #         raise KeyError("Cannot update item with ID '%s': "
    #                        "ID does not exist")
    #     return item

    # def delete_item(self, id_, context=None):
        # """Remove item from model.

        # Args:
        #     id_: The ID of the item to be removed
        #     context: Key-values providing frame of reference of request

        # Returns:
        #      The removed item.

        # Raises:
        #     KeyError: Item with specified id_ not present.
        # """



# Copyright (c) 2012 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import datetime

from oslo_db.sqlalchemy import models
from oslo_utils import uuidutils
import sqlalchemy as sa
from sqlalchemy.ext import declarative
from sqlalchemy import orm


class CongressBase(models.ModelBase):
    """Base class for Congress Models."""

    __table_args__ = {'mysql_engine': 'InnoDB'}

    @declarative.declared_attr
    def __tablename__(cls):
        # NOTE(jkoelker) use the pluralized name of the class as the table
        return cls.__name__.lower() + 's'

    def __iter__(self):
        self._i = iter(orm.object_mapper(self).columns)
        return self

    def next(self):
        n = next(self._i).name
        return n, getattr(self, n)

    def __repr__(self):
        """sqlalchemy based automatic __repr__ method."""
        items = ['%s=%r' % (col.name, getattr(self, col.name))
                 for col in self.__table__.columns]
        return "<%s.%s[object at %x] {%s}>" % (self.__class__.__module__,
                                               self.__class__.__name__,
                                               id(self), ', '.join(items))


class HasTenant(object):
    """Tenant mixin, add to subclasses that have a tenant."""

    # NOTE(jkoelker) tenant_id is just a free form string ;(
    tenant_id = sa.Column(sa.String(255))


class HasId(object):
    """id mixin, add to subclasses that have an id."""

    id = sa.Column(sa.String(36),
                   primary_key=True,
                   default=uuidutils.generate_uuid)


# TODO(arosen) move this somewhere better later...
def _get_date():
    return datetime.datetime.now()


class HasAudit(object):
    created_at = sa.Column(sa.DateTime, default=_get_date, nullable=False)
    updated_at = sa.Column(sa.DateTime, onupdate=_get_date, nullable=True)
    deleted_at = sa.Column(sa.DateTime, nullable=True)
    deleted = sa.Column(sa.String(length=36),
                        server_default='', default='', nullable=True)


BASE = declarative.declarative_base(cls=CongressBase)



# Copyright (c) 2015 Cisco.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import datetime

import keystoneclient.v3.client as ksclient
from monascaclient import client as monasca_client
from oslo_log import log as logging
from oslo_utils import timeutils

from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils

LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    """Create a dataservice instance.

    This method is called by d6cage to create a dataservice
    instance.  There are a couple of parameters we found useful
    to add to that call, so we included them here instead of
    modifying d6cage (and all the d6cage.createservice calls).
    """
    return MonascaDriver(name, keys, inbox, datapath, args)


# TODO(thinrichs): figure out how to move even more of this boilerplate
#   into DataSourceDriver.  E.g. change all the classes to Driver instead of
#   NeutronDriver, CeilometerDriver, etc. and move the d6instantiate function
#   to DataSourceDriver.
class MonascaDriver(datasource_driver.PollingDataSourceDriver,
                    datasource_driver.ExecutionDriver):

    METRICS = "metrics"
    DIMENSIONS = "dimensions"
    STATISTICS = "statistics"
    DATA = "statistics.data"
    # TODO(fabiog): add events and logs when fully supported in Monasca
    # EVENTS = "events"
    # LOGS = "logs"

    value_trans = {'translation-type': 'VALUE'}

    metric_translator = {
        'translation-type': 'HDICT',
        'table-name': METRICS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'dimensions',
              'translator': {'translation-type': 'VDICT',
                             'table-name': DIMENSIONS,
                             'id-col': 'id',
                             'key-col': 'key', 'val-col': 'value',
                             'translator': value_trans}})
    }

    statistics_translator = {
        'translation-type': 'HDICT',
        'table-name': STATISTICS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'statistics',
              'translator': {'translation-type': 'LIST',
                             'table-name': DATA,
                             'id-col': 'name',
                             'val-col': 'value_col',
                             'translator': value_trans}})
    }

    TRANSLATORS = [metric_translator, statistics_translator]

    def __init__(self,  name='', keys='', inbox=None, datapath=None,
                 args=None):
        super(MonascaDriver, self).__init__(name, keys, inbox, datapath, args)
        datasource_driver.ExecutionDriver.__init__(self)
        self.creds = args
        if not self.creds.get('project_name'):
            self.creds['project_name'] = self.creds['tenant_name']

        if not self.creds.get('poll_time'):
            # set default polling time to 1hr
            self.creds['poll_time'] = 3600

        # Monasca uses Keystone V3
        self.creds['auth_url'] = self.creds['auth_url'].replace("v2.0", "v3")
        self.keystone = ksclient.Client(**self.creds)
        self.creds['token'] = self.keystone.auth_token

        if not self.creds.get('endpoint'):
            # if the endpoint not defined retrieved it from keystone catalog
            self.creds['endpoint'] = self.keystone.service_catalog.url_for(
                service_type='monitoring', endpoint_type='publicURL')

        self.monasca = monasca_client.Client('2_0', **self.creds)
        self.add_executable_client_methods(self.monasca, 'monascaclient.')
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'monasca'
        result['description'] = ('Datasource driver that interfaces with '
                                 'monasca.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['secret'] = ['password']
        return result

    def update_from_datasource(self):
        """Read Data from Monasca datasource.

        And to fill up the current state of the policy engine.
        """

        try:
            LOG.debug("Monasca grabbing metrics")
            metrics = self.monasca.metrics.list()
            self._translate_metric(metrics)
            LOG.debug("METRICS: %s" % str(self.state[self.METRICS]))

            LOG.debug("Monasca grabbing statistics")
            # gather statistic for the last day
            today = datetime.datetime.now()
            yesterday = datetime.timedelta(hours=24)
            start_from = timeutils.isotime(today-yesterday)

            for metric in self.monasca.metrics.list_names():
                LOG.debug("Monasca statistics for metric %s", metric['name'])
                _query_args = dict(
                    start_time=start_from,
                    name=metric['name'],
                    statistics='avg',
                    period=int(self.creds['poll_time']),
                    merge_metrics='true')
                statistics = self.monasca.metrics.list_statistics(
                    **_query_args)
                self._translate_statistics(statistics)
                LOG.debug("STATISTICS: %s" % str(self.state[self.STATISTICS]))

        except Exception as e:
            raise e

    @ds_utils.update_state_on_changed(METRICS)
    def _translate_metric(self, obj):
        """Translate the metrics represented by OBJ into tables."""
        LOG.debug("METRIC: %s" % str(obj))

        row_data = MonascaDriver.convert_objs(obj,
                                              self.metric_translator)
        return row_data

    @ds_utils.update_state_on_changed(STATISTICS)
    def _translate_statistics(self, obj):
        """Translate the metrics represented by OBJ into tables."""

        LOG.debug("STATISTICS: %s" % str(obj))

        row_data = MonascaDriver.convert_objs(obj,
                                              self.statistics_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.monasca, action, action_args)



# Copyright (c) 2013 VMware, Styra.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import mock
import novaclient
import time

from oslo_config import cfg
cfg.CONF.distributed_architecture = True
from oslo_messaging import conffixture

from congress.datalog import compile
from congress.datasources.nova_driver import NovaDriver
from congress import exception as congressException
from congress.policy_engines.agnostic import Dse2Runtime
from congress.tests import base
from congress.tests.fake_datasource import FakeDataSource
from congress.tests import helper


class TestDSE(base.TestCase):

    def setUp(self):
        super(TestDSE, self).setUp()
        mc_fixture = conffixture.ConfFixture(cfg.CONF)
        mc_fixture.conf.transport_url = 'kombu+memory://'
        self.messaging_config = mc_fixture.conf
        self.messaging_config.rpc_response_timeout = 1

    def test_intranode_pubsub(self):
        node = helper.make_dsenode_new_partition('testnode')
        test1 = FakeDataSource('test1')
        test2 = FakeDataSource('test2')
        node.register_service(test1)
        node.register_service(test2)

        test1.subscribe('test2', 'p')
        helper.retry_check_function_return_value(
            lambda: hasattr(test1, 'last_msg'), True)
        test2.publish('p', 42)
        helper.retry_check_function_return_value(
            lambda: test1.last_msg['data'], 42)
        self.assertFalse(hasattr(test2, "last_msg"))

    def test_intranode_pubsub2(self):
        # same as test_intranode_pubsub but with opposite ordering.
        # (Ordering does matter with internode_pubsub).
        node = helper.make_dsenode_new_partition('testnode')
        test1 = FakeDataSource('test1')
        test2 = FakeDataSource('test2')
        node.register_service(test1)
        node.register_service(test2)

        test2.subscribe('test1', 'p')
        helper.retry_check_function_return_value(
            lambda: hasattr(test2, 'last_msg'), True)
        test1.publish('p', 42)
        helper.retry_check_function_return_value(
            lambda: test2.last_msg['data'], 42)
        self.assertFalse(hasattr(test1, "last_msg"))

    def test_intranode_partial_unsub(self):
        node = helper.make_dsenode_new_partition('testnode')
        test1 = FakeDataSource('test1')
        test2 = FakeDataSource('test2')
        node.register_service(test1)
        node.register_service(test2)

        test1.subscribe('test2', 'p')
        test1.subscribe('test2', 'q')
        test1.unsubscribe('test2', 'q')  # unsub from q should not affect p
        helper.retry_check_function_return_value(
            lambda: hasattr(test1, 'last_msg'), True)
        test2.publish('p', 42)
        helper.retry_check_function_return_value(
            lambda: test1.last_msg['data'], 42)
        self.assertFalse(hasattr(test2, "last_msg"))

    def test_internode_pubsub(self):
        node1 = helper.make_dsenode_new_partition('testnode1')
        test1 = FakeDataSource('test1')
        node1.register_service(test1)
        node2 = helper.make_dsenode_same_partition(node1, 'testnode2')
        test2 = FakeDataSource('test2')
        node2.register_service(test2)

        test1.subscribe('test2', 'p')
        helper.retry_check_function_return_value(
            lambda: hasattr(test1, 'last_msg'), True)
        test2.publish('p', 42)
        helper.retry_check_function_return_value(
            lambda: test1.last_msg['data'], 42)
        self.assertFalse(hasattr(test2, "last_msg"))

    def test_internode_partial_unsub(self):
        node1 = helper.make_dsenode_new_partition('testnode1')
        node2 = helper.make_dsenode_same_partition(node1, 'testnode2')
        test1 = FakeDataSource('test1')
        test2 = FakeDataSource('test2')
        node1.register_service(test1)
        node2.register_service(test2)

        test1.subscribe('test2', 'p')
        test1.subscribe('test2', 'q')
        test1.unsubscribe('test2', 'q')  # unsub from q should not affect p
        helper.retry_check_function_return_value(
            lambda: hasattr(test1, 'last_msg'), True)
        test2.publish('p', 42)
        helper.retry_check_function_return_value(
            lambda: test1.last_msg['data'], 42)
        self.assertFalse(hasattr(test2, "last_msg"))

    def test_multiservice_pubsub(self):
        node1 = helper.make_dsenode_new_partition('testnode1')
        test1 = FakeDataSource('test1')
        test2 = FakeDataSource('test2')
        node1.register_service(test1)
        node1.register_service(test2)
        node2 = helper.make_dsenode_same_partition(node1, 'testnode2')
        test3 = FakeDataSource('test3')
        node2.register_service(test3)

        test1.subscribe('test3', 'p')
        helper.retry_check_function_return_value(
            lambda: hasattr(test1, 'last_msg'), True)
        test3.publish('p', 42)
        helper.retry_check_function_return_value(
            lambda: test1.last_msg['data'], 42)
        self.assertFalse(hasattr(test2, "last_msg"))
        self.assertFalse(hasattr(test3, "last_msg"))

    def test_subscribe_snapshot(self):
        node = helper.make_dsenode_new_partition('testnode')
        test1 = FakeDataSource('test1')
        test2 = FakeDataSource('test2')
        node.register_service(test1)
        node.register_service(test2)

        test1.subscribe('test2', 'fake_table')
        helper.retry_check_function_return_value(
            lambda: hasattr(test1, 'last_msg'), True)
        self.assertEqual(test1.last_msg['data'], test2.state['fake_table'])

    def test_datasource_sub(self):
        node = helper.make_dsenode_new_partition('testnode')
        nova_client = mock.MagicMock()
        with mock.patch.object(novaclient.client.Client, '__init__',
                               return_value=nova_client):
            nova = NovaDriver(
                name='nova', args=helper.datasource_openstack_args())
            test = FakeDataSource('test')
            node.register_service(nova)
            node.register_service(test)

            nova.subscribe('test', 'p')
            helper.retry_check_function_return_value(
                lambda: hasattr(nova, 'last_msg'), True)
            test.publish('p', 42)
            helper.retry_check_function_return_value(
                lambda: nova.last_msg['data'], 42)
            self.assertFalse(hasattr(test, "last_msg"))

    def test_datasource_unsub(self):
        node = helper.make_dsenode_new_partition('testnode')
        nova_client = mock.MagicMock()
        with mock.patch.object(novaclient.client.Client, '__init__',
                               return_value=nova_client):
            nova = NovaDriver(
                name='nova', args=helper.datasource_openstack_args())
            test = FakeDataSource('test')
            node.register_service(nova)
            node.register_service(test)

            nova.subscribe('test', 'p')
            helper.retry_check_function_return_value(
                lambda: hasattr(nova, 'last_msg'), True)
            test.publish('p', 42)
            helper.retry_check_function_return_value(
                lambda: nova.last_msg['data'], 42)
            self.assertFalse(hasattr(test, "last_msg"))
            nova.unsubscribe('test', 'p')
            test.publish('p', 43)
            # hard to test that the message is never delivered
            time.sleep(0.2)
            self.assertEqual(nova.last_msg['data'], 42)

    def test_datasource_pub(self):
        node = helper.make_dsenode_new_partition('testnode')
        nova_client = mock.MagicMock()
        with mock.patch.object(novaclient.client.Client, '__init__',
                               return_value=nova_client):
            nova = NovaDriver(
                name='nova', args=helper.datasource_openstack_args())
            test = FakeDataSource('test')
            node.register_service(nova)
            node.register_service(test)

            test.subscribe('nova', 'p')
            helper.retry_check_function_return_value(
                lambda: hasattr(test, 'last_msg'), True)
            nova.publish('p', 42)
            helper.retry_check_function_return_value(
                lambda: test.last_msg['data'], 42)
            self.assertFalse(hasattr(nova, "last_msg"))

    def test_datasource_poll(self):
        node = helper.make_dsenode_new_partition('testnode')
        pub = FakeDataSource('pub')
        sub = FakeDataSource('sub')
        node.register_service(pub)
        node.register_service(sub)

        sub.subscribe('pub', 'fake_table')
        pub.state = {'fake_table': set([(1, 2)])}
        pub.poll()
        helper.retry_check_function_return_value(
            lambda: sub.last_msg['data'], set(pub.state['fake_table']))
        self.assertFalse(hasattr(pub, "last_msg"))

    def test_policy(self):
        node = helper.make_dsenode_new_partition('testnode')
        data = FakeDataSource('data')
        engine = Dse2Runtime('engine')
        node.register_service(data)
        node.register_service(engine)

        engine.create_policy('alpha')
        engine.create_policy('data')
        self.insert_rule(engine, 'p(x) :- data:fake_table(x)', 'alpha')
        data.state = {'fake_table': set([(1,), (2,)])}
        data.poll()
        helper.retry_check_db_equal(
            engine, 'p(x)', 'p(1) p(2)', target='alpha')
        self.assertFalse(hasattr(engine, "last_msg"))

    def insert_rule(self, engine, statement, target=None):
        statement = compile.parse1(statement)
        if target is None:
            e = compile.Event(statement)
        else:
            e = compile.Event(statement, target=target)
        engine.process_policy_update([e])

    def test_unregister(self):
        node = helper.make_dsenode_new_partition('testnode')
        test1 = FakeDataSource('test1')
        node.register_service(test1)
        obj = node.invoke_service_rpc(
            'test1', 'get_status', source_id=None, params=None)
        self.assertIsNotNone(obj)
        node.unregister_service('test1')
        helper.retry_til_exception(
            congressException.NotFound,
            lambda: node.invoke_service_rpc(
                'test1', 'get_status', source_id=None, params=None))

    def _create_node_with_services(self, nodes, services, num, partition_id):
        nid = 'cbd_node%s' % num
        nodes.append(helper.make_dsenode_same_partition(partition_id, nid))
        ns = []
        for s in range(num):
            # intentionally starting different number services
            ns.append(FakeDataSource('cbd-%d_svc-%d' % (num, s)))
            nodes[-1].register_service(ns[-1])
        services.append(ns)
        return nodes[-1]

    def test_subs_list_update_aggregated_by_service(self):
        part = helper.get_new_partition()
        nodes = []
        services = []
        num_nodes = 3

        for i in range(num_nodes):
            n = self._create_node_with_services(nodes, services, i, part)
            n.start()

        # add subscriptions
        for i in range(2, num_nodes):
            for s2 in services[i]:
                for s1 in services[i-1]:
                    s1.subscribe(s2.service_id, 'table-A')
                    s2.subscribe(s1.service_id, 'table-B')
        services[1][0].subscribe(services[2][0].service_id, 'table-C')
        services[2][1].subscribe(services[2][0].service_id, 'table-D')

        # constructed expected results
        expected_subbed_tables = {}
        expected_subbed_tables[nodes[1].node_id] = {}
        expected_subbed_tables[nodes[2].node_id] = {}
        expected_subbed_tables[nodes[1].node_id][
            services[1][0].service_id] = set(['table-B'])
        expected_subbed_tables[nodes[2].node_id][
            services[2][0].service_id] = set(['table-A', 'table-C', 'table-D'])
        expected_subbed_tables[nodes[2].node_id][
            services[2][1].service_id] = set(['table-A'])

        # validate
        def _validate_subbed_tables(node):
            for s in node.get_services():
                sid = s.service_id
                subscribed_tables = node.service_object(
                    sid)._published_tables_with_subscriber
                self.assertEqual(
                    subscribed_tables,
                    expected_subbed_tables[node.node_id][sid],
                    '%s has incorrect subscribed tables list' % sid)
            return True
        for n in nodes:
            helper.retry_check_function_return_value(
                lambda: _validate_subbed_tables(n), True)

        # selectively unsubscribe
        services[1][0].unsubscribe(services[2][0].service_id, 'table-A')
        # note that services[2][1] still subscribes to 'table-B'
        services[2][0].unsubscribe(services[1][0].service_id, 'table-B')
        # extraneous unsubscribe
        services[2][0].unsubscribe(services[1][0].service_id, 'table-None')

        # update expected results
        expected_subbed_tables[nodes[2].node_id][
            services[2][0].service_id] = set(['table-C', 'table-D'])

        for n in nodes:
            helper.retry_check_function_return_value(
                lambda: _validate_subbed_tables(n), True)

        # resubscribe
        services[1][0].subscribe(services[2][0].service_id, 'table-A')
        services[2][0].subscribe(services[1][0].service_id, 'table-B')

        # update expected results
        expected_subbed_tables[nodes[2].node_id][
            services[2][0].service_id] = set(['table-A', 'table-C', 'table-D'])

        for n in nodes:
            helper.retry_check_function_return_value(
                lambda: _validate_subbed_tables(n), True)



# Copyright (c) 2013 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging

from congress.datalog import utility
from congress.tests import base

LOG = logging.getLogger(__name__)


class TestOrderedSet(base.TestCase):
    def test_creation_simple(self):
        """"Test basic OrderedSet instantiation."""
        contents = ["foo", "bar", "baz"]
        os = utility.OrderedSet(contents)
        self.assertEqual(list(os), contents)
        self.assertEqual(len(os), len(contents))
        self.assertEqual(set(os), set(contents))

    def test_creation_with_duplicates(self):
        """"Test that OrderedSet instantiation removes duplicates."""
        contents = ["foo", "bar", "foo", "baz"]
        os = utility.OrderedSet(contents)
        self.assertNotEqual(list(os), contents)
        self.assertEqual(len(os), len(contents) - 1)
        self.assertEqual(set(os), set(contents))

    def test_contains(self):
        """Test that basic OrderedSet.__contains__ functionality works."""
        contents = ["foo", "bar", "baz"]
        missing = "qux"
        os = utility.OrderedSet(contents)
        self.assertTrue(all(x in os for x in contents))
        self.assertTrue(missing not in os)

        discarded = contents[1]
        os.discard(discarded)
        self.assertTrue(all(x in os for x in contents if x != discarded))
        self.assertTrue(discarded not in os)

    def test_add_known_item(self):
        """Test that OrderedSet.add(known) returns False."""
        contents = ["foo", "bar", "baz"]
        known = contents[1]
        os = utility.OrderedSet(contents)
        self.assertFalse(os.add(known))
        self.assertEqual(list(os), contents)

    def test_add_unknown_item(self):
        """Test that OrderedSet.add(unknown) returns True."""
        contents = ["foo", "bar", "baz"]
        unknown = "qux"
        os = utility.OrderedSet(contents)
        self.assertTrue(os.add(unknown))
        self.assertEqual(list(os), contents + [unknown])

    def test_discard_known_item(self):
        """Test that OrderedSet.discard(known) returns True."""
        contents = ["foo", "bar", "baz"]
        known = contents[1]
        new_contents = [x for x in contents if x != known]
        os = utility.OrderedSet(contents)
        self.assertTrue(os.discard(known))
        self.assertEqual(list(os), new_contents)

    def test_discard_unknown_item(self):
        """Test that OrderedSet.discard(unknown) returns False."""
        contents = ["foo", "bar", "baz"]
        unknown = "qux"
        os = utility.OrderedSet(contents)
        self.assertFalse(os.discard(unknown))
        self.assertEqual(list(os), contents)

    def test_pop_last_item(self):
        """Test that OrderedSet.pop() returns the final item."""
        contents = ["foo", "bar", "baz"]
        os = utility.OrderedSet(contents)
        self.assertEqual(os.pop(), contents[-1])
        self.assertEqual(list(os), contents[:-1])

    def test_pop_not_first_item(self):
        """Test that OrderedSet.pop(last=False) returns the first item."""
        contents = ["foo", "bar", "baz"]
        os = utility.OrderedSet(contents)
        self.assertEqual(os.pop(last=False), contents[0])
        self.assertEqual(list(os), contents[1:])

    def test_reversed_reverses_order(self):
        """Test that reversed(OrderedSet()) reverses correctly."""
        contents = ["foo", "bar", "baz"]
        os = utility.OrderedSet(contents)
        self.assertEqual(list(reversed(os)), list(reversed(contents)))

    def test_equals_other_ordered_set(self):
        """Test that OrderedSet equality accounts for order."""
        contents = ["foo", "bar", "baz"]
        os = utility.OrderedSet(contents)
        self.assertNotEqual(os, utility.OrderedSet(reversed(os)))
        self.assertEqual(os, utility.OrderedSet(contents))

    def test_equals_other_iterable(self):
        """Test that OrderedSet-to-other-iterable equality returns False."""
        contents = ["foo", "bar", "baz"]
        os = utility.OrderedSet(contents)
        self.assertNotEqual(os, set(os))
        self.assertNotEqual(os, frozenset(os))
        self.assertNotEqual(os, list(os))
        self.assertNotEqual(os, tuple(os))
        self.assertNotEqual(os, {x: 0 for x in os})



# Copyright (c) 2014 Montavista Software, LLC.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import cinderclient.client

from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils


def d6service(name, keys, inbox, datapath, args):
    """This method is called by d6cage to create a dataservice instance."""
    return CinderDriver(name, keys, inbox, datapath, args)


class CinderDriver(datasource_driver.PollingDataSourceDriver,
                   datasource_driver.ExecutionDriver):
    VOLUMES = "volumes"
    SNAPSHOTS = "snapshots"
    SERVICES = "services"

    # This is the most common per-value translator, so define it once here.
    value_trans = {'translation-type': 'VALUE'}

    volumes_translator = {
        'translation-type': 'HDICT',
        'table-name': VOLUMES,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'translator': value_trans},
             {'fieldname': 'size', 'translator': value_trans},
             {'fieldname': 'user_id', 'translator': value_trans},
             {'fieldname': 'status', 'translator': value_trans},
             {'fieldname': 'description', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'bootable', 'translator': value_trans},
             {'fieldname': 'created_at', 'translator': value_trans},
             {'fieldname': 'volume_type', 'translator': value_trans})}

    snapshots_translator = {
        'translation-type': 'HDICT',
        'table-name': SNAPSHOTS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'translator': value_trans},
             {'fieldname': 'size', 'translator': value_trans},
             {'fieldname': 'status', 'translator': value_trans},
             {'fieldname': 'volume_id', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'created_at', 'translator': value_trans})}

    services_translator = {
        'translation-type': 'HDICT',
        'table-name': SERVICES,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'status', 'translator': value_trans},
             {'fieldname': 'binary', 'translator': value_trans},
             {'fieldname': 'zone', 'translator': value_trans},
             {'fieldname': 'state', 'translator': value_trans},
             {'fieldname': 'updated_at', 'translator': value_trans},
             {'fieldname': 'host', 'translator': value_trans},
             {'fieldname': 'disabled_reason', 'translator': value_trans})}

    TRANSLATORS = [volumes_translator, snapshots_translator,
                   services_translator]

    def __init__(self, name='', keys='', inbox=None, datapath=None, args=None):
        super(CinderDriver, self).__init__(name, keys, inbox, datapath, args)
        datasource_driver.ExecutionDriver.__init__(self)
        self.creds = self.get_cinder_credentials_v2(args)
        self.cinder_client = cinderclient.client.Client(**self.creds)
        self.add_executable_client_methods(self.cinder_client,
                                           'cinderclient.v2.')
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'cinder'
        result['description'] = ('Datasource driver that interfaces with '
                                 'OpenStack cinder.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['secret'] = ['password']
        return result

    def update_from_datasource(self):
        volumes = self.cinder_client.volumes.list(
            detailed=True, search_opts={"all_tenants": 1})
        self._translate_volumes(volumes)

        snapshots = self.cinder_client.volume_snapshots.list(
            detailed=True, search_opts={"all_tenants": 1})
        self._translate_snapshots(snapshots)

        services = self.cinder_client.services.list(
            host=None, binary=None)
        self._translate_services(services)

    def get_cinder_credentials_v2(self, creds):
        d = {}
        d['version'] = '2'
        d['username'] = creds['username']
        d['api_key'] = creds['password']
        d['auth_url'] = creds['auth_url']
        d['project_id'] = creds['tenant_name']
        return d

    @ds_utils.update_state_on_changed(VOLUMES)
    def _translate_volumes(self, obj):
        row_data = CinderDriver.convert_objs(obj, self.volumes_translator)
        return row_data

    @ds_utils.update_state_on_changed(SNAPSHOTS)
    def _translate_snapshots(self, obj):
        row_data = CinderDriver.convert_objs(obj, self.snapshots_translator)
        return row_data

    @ds_utils.update_state_on_changed(SERVICES)
    def _translate_services(self, obj):
        row_data = CinderDriver.convert_objs(obj, self.services_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.cinder_client, action, action_args)



# Copyright (c) 2015 Intel, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from congress.api import api_utils
from congress.api import base
from congress.api import webservice
from congress import exception


def d6service(name, keys, inbox, datapath, args):
    return ActionsModel(name, keys, inbox=inbox, dataPath=datapath, **args)


class ActionsModel(base.APIModel):
    """Model for handling API requests about Actions."""
    def __init__(self, name, keys='', inbox=None, dataPath=None,
                 policy_engine=None, datasource_mgr=None):
        super(ActionsModel, self).__init__(name, keys, inbox=inbox,
                                           dataPath=dataPath,
                                           policy_engine=policy_engine,
                                           datasource_mgr=datasource_mgr)

    def get_items(self, params, context=None):
        """Retrieve items from this model.

        Args:
            params: A dict-like object containing parameters
                    from the request query string and body.
            context: Key-values providing frame of reference of request

        Returns:
             A dict containing at least a 'actions' key whose value is a list
             of items in this model.
        """
        caller, source_id = api_utils.get_id_from_context(
            context, self.datasource_mgr, self.engine)

        try:
            rpc_args = {'source_id': source_id}
            return self.invoke_rpc(caller, 'get_actions', rpc_args)
        except exception.CongressException as e:
            raise webservice.DataModelException(
                exception.NotFound.code, str(e),
                http_status_code=exception.NotFound.code)
